# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# The LocalDriver provides the following:
#
# 1- A getpackages function based on finding all directories inside a specific
#    git repo, where each directory represents a package
#
# 2- A getinfo function based on a single-distgit repo paradigm

import logging
import os
import sh
import shutil

from dlrn.db import Commit
from dlrn.drivers.pkginfo import PkgInfoDriver
from rdopkg.utils import specfile

logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger("dlrn-local-driver")
logger.setLevel(logging.INFO)


class LocalDriver(PkgInfoDriver):
    DRIVER_CONFIG = {
        'local_driver': {
        }
    }

    def _get_infos_from_pkg(self, packagepath):
        version = None
        package = None
        if not os.path.isdir(os.path.join(packagepath, '.git')):
            raise RuntimeError(
                'The distgit directory must a be git repository')
        for pkgfile in os.listdir(packagepath):
            if pkgfile.endswith('.spec'):
                spec = specfile.Spec(fn=os.path.join(packagepath, pkgfile))
                version = spec.get_tag('Version')
                package = pkgfile.replace('.spec', '')
        if not version:
            raise RuntimeError("No spec file or no version found")
        return package, version

    def getpackages(self, **kwargs):
        datadir = self.config_options.datadir
        # src_dir is only set for the test case. Normal behavior is
        # to look at the current directory.
        if 'src_dir' in kwargs:
            src_dir = kwargs['src_dir']
        else:
            src_dir = os.getcwd()
        is_src_dir_distgit = [
            f for f in os.listdir(src_dir) if f.endswith('.spec')]
        if not is_src_dir_distgit:
            return []
        package, version = self._get_infos_from_pkg(src_dir)
        dest_dir = os.path.join(datadir, package)
        logger.info("Copy distgit source from %s to %s" % (src_dir, dest_dir))
        if os.path.isdir(dest_dir):
            shutil.rmtree(dest_dir)
        shutil.copytree(src_dir, dest_dir)
        pkg_hash = {}
        pkg_hash['name'] = package
        pkg_hash['maintainers'] = 'test@example.com'
        pkg_hash['master-distgit'] = dest_dir
        pkg_hash['upstream'] = 'Unknown'
        logger.info(
            "Got version %s for %s from the spec" % (version, package))
        pkg_hash['source-branch'] = version
        return [pkg_hash]

    def getinfo(self, **kwargs):
        project = kwargs.get('project')
        package = kwargs.get('package')
        datadir = self.config_options.datadir
        distro_dir = os.path.join(datadir, package['name'])

        # Get distro_hash from last commit in distgit directory
        git = sh.git.bake(_cwd=package['master-distgit'], _tty_out=False)
        repoinfo = str(git.log("--pretty=format:%H %ct", "-1", ".")
                       ).strip().split(" ")
        distro_hash = repoinfo[0]
        dt_distro = repoinfo[1]

        project_toprocess = []
        commit_hash = "%s-%s" % (project, package['source-branch'])
        commit = Commit(
            dt_commit=0, project_name=project,
            type='rpm',
            commit_hash=commit_hash, repo_dir=None,
            distro_hash=distro_hash, dt_distro=dt_distro,
            distgit_dir=distro_dir,
            commit_branch=package['source-branch'],
            dt_extended=0, extended_hash=None, component=None)
        project_toprocess.append(commit)

        return project_toprocess

    def preprocess(self, **kwargs):
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add user column to civote and promotion

Revision ID: cab7697f6564
Revises: 4a5651777e5e
Create Date: 2017-11-24 13:14:05.750269

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'cab7697f6564'
down_revision = '4a5651777e5e'
branch_labels = None
depends_on = None


def upgrade():
    # Note that existing votes and promotions will have "null" as the user
    # Since SQLite3 does not allow ALTER TABLE statements, we need to do a
    # batch operation: http://alembic.zzzcomputing.com/en/latest/batch.html
    with op.batch_alter_table('civotes') as batch_op:
        batch_op.add_column(sa.Column('user',
                                      sa.String(256),
                                      sa.ForeignKey('users.username',
                                                    name='civ_user_fk')))
    with op.batch_alter_table('promotions') as batch_op:
        batch_op.add_column(sa.Column('user', sa.String(256),
                            sa.ForeignKey('users.username',
                                          name='prom_user_fk')))


def downgrade():
    with op.batch_alter_table('civotes') as batch_op:
        batch_op.drop_constraint('civ_user_fk', type_='foreignkey')
        batch_op.drop_column('user')

    with op.batch_alter_table('promotions') as batch_op:
        batch_op.drop_constraint('prom_user_fk', type_='foreignkey')
        batch_op.drop_column('user')



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import sh
import shutil
import tempfile

from six.moves import configparser

from dlrn.config import ConfigOptions
from dlrn.drivers.local import LocalDriver
from dlrn.tests import base


class TestDriverLocal(base.TestCase):
    def setUp(self):
        super(TestDriverLocal, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set("DEFAULT", "pkginfo_driver",
                   "dlrn.drivers.local.LocalDriver")
        self.config = ConfigOptions(config)
        self.config.datadir = tempfile.mkdtemp()
        self.base_dir = tempfile.mkdtemp()

    def tearDown(self):
        super(TestDriverLocal, self).tearDown()
        shutil.rmtree(self.config.datadir)
        shutil.rmtree(self.base_dir)

    @mock.patch.object(sh.Command, '__call__', autospec=True)
    def test_getinfo(self, sh_mock):
        driver = LocalDriver(cfg_options=self.config)
        package = {
            'upstream': 'Unknown', 'name': 'hello-world',
            'master-distgit': '/tmp/hello-world', 'source-branch': '1.0.0'}
        info = driver.getinfo(
            package=package, project="hello-world")
        self.assertEqual(len(info), 1)
        self.assertEqual(info[0].commit_hash, 'hello-world-1.0.0')
        self.assertEqual(info[0].commit_branch, '1.0.0')

    def test_getpackages(self):
        package = 'hello-world'
        distgit_dir = os.path.join(self.base_dir, package)
        os.mkdir(distgit_dir)
        os.mkdir(os.path.join(distgit_dir, '.git'))
        open(os.path.join(distgit_dir, 'hello-world.spec'), 'w').write(
            'Version: 1.0.0'
        )
        # We expect to call DLRN from the local distgit directory
        driver = LocalDriver(cfg_options=self.config)
        packages = driver.getpackages(src_dir=distgit_dir)
        self.assertEqual(len(packages), 1)
        expected_package = {
            'maintainers': 'test@example.com',
            'master-distgit': '%s/hello-world' % self.config.datadir,
            'name': 'hello-world',
            'source-branch': '1.0.0',
            'upstream': 'Unknown'}
        self.assertDictEqual(packages[0], expected_package)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import shutil
import sys
import tempfile

from dlrn.config import ConfigOptions
from dlrn import db
from dlrn import shell
from dlrn.tests import base
from dlrn import utils

from six.moves import configparser


def mocked_session(url):
    session = db.getSession(url)
    utils.loadYAML(session, './dlrn/tests/samples/commits_1.yaml')
    return session


def mocked_session_recheck(url):
    db_fd, filepath = tempfile.mkstemp()
    session = db.getSession("sqlite:///%s" % filepath)
    utils.loadYAML(session, './dlrn/tests/samples/commits_1.yaml')
    return session


def mocked_getpackages(**kwargs):
    return [{'upstream': 'https://github.com/openstack/foo',
             'name': 'foo', 'maintainers': 'test@test.com'},
            {'upstream': 'https://github.com/openstack/test',
             'name': 'test', 'maintainers': 'test@test.com'},
            {'upstream': 'https://github.com/openstack/test',
             'name': 'python-pysaml2', 'maintainers': 'test@test.com'}]


class TestProcessBuildResult(base.TestCase):
    def setUp(self):
        super(TestProcessBuildResult, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set('DEFAULT', 'datadir', tempfile.mkdtemp())
        config.set('DEFAULT', 'scriptsdir', tempfile.mkdtemp())
        config.set('DEFAULT', 'baseurl', "file://%s" % config.get('DEFAULT',
                                                                  'datadir'))
        self.config = ConfigOptions(config)
        self.commit = db.Commit(dt_commit=123, project_name='foo', type="rpm",
                                commit_hash='1c67b1ab8c6fe273d4e175a14f0df5'
                                            'd3cbbd0edf',
                                repo_dir='/home/dlrn/data/foo',
                                distro_hash='c31d1b18eb5ab5aed6721fc4fad06c9'
                                            'bd242490f',
                                dt_distro=123,
                                distgit_dir='/home/dlrn/data/foo_distro',
                                commit_branch='master', dt_build=1441245153)
        self.db_fd, filepath = tempfile.mkstemp()
        self.session = mocked_session("sqlite:///%s" % filepath)
        self.packages = [{'upstream': 'https://github.com/openstack/foo',
                          'name': 'foo', 'maintainers': 'test@test.com'},
                         {'upstream': 'https://github.com/openstack/test',
                          'name': 'test', 'maintainers': 'test@test.com'}]

    def tearDown(self):
        super(TestProcessBuildResult, self).tearDown()
        shutil.rmtree(self.config.datadir)
        shutil.rmtree(self.config.scriptsdir)
        os.close(self.db_fd)

    @mock.patch('os.rename')
    @mock.patch('os.symlink')
    @mock.patch('dlrn.shell.export_commit_yaml')
    @mock.patch('dlrn.shell.genreports')
    @mock.patch('dlrn.shell.sync_repo')
    def test_successful_build(self, rs_mock, gr_mock, ec_mock, sl_mock,
                              rn_mock):
        built_rpms = ['foo-1.2.3.rpm']
        status = [self.commit, built_rpms, 'OK', None]
        output = shell.process_build_result(status, self.packages,
                                            self.session, [])
        self.assertEqual(output, 0)
        self.assertEqual(gr_mock.call_count, 1)
        self.assertEqual(rs_mock.call_count, 1)
        self.assertEqual(ec_mock.call_count, 1)
        self.assertEqual(sl_mock.call_count, 1)
        self.assertEqual(rn_mock.call_count, 1)

    @mock.patch('dlrn.shell.export_commit_yaml')
    @mock.patch('dlrn.shell.sendnotifymail')
    @mock.patch('dlrn.shell.genreports')
    @mock.patch('dlrn.shell.sync_repo')
    def test_failed_build(self, rs_mock, gr_mock, sm_mock, ec_mock):
        error_msg = 'Unit test error'
        status = [self.commit, '', '', error_msg]
        output = shell.process_build_result(status, self.packages,
                                            self.session, [])
        self.assertEqual(output, 1)
        self.assertEqual(gr_mock.call_count, 1)
        self.assertEqual(rs_mock.call_count, 1)
        self.assertEqual(sm_mock.call_count, 1)
        self.assertEqual(ec_mock.call_count, 1)

    @mock.patch('dlrn.shell.submit_review')
    @mock.patch('dlrn.shell.sendnotifymail')
    @mock.patch('dlrn.shell.genreports')
    @mock.patch('dlrn.shell.sync_repo')
    def test_failed_build_gerrit(self, rs_mock, gr_mock, sm_mock, sr_mock):
        self.config.gerrit = 'yes'
        error_msg = 'Unit test error'
        status = [self.commit, '', '', error_msg]
        output = shell.process_build_result(status, self.packages,
                                            self.session, [])
        self.assertEqual(output, 1)
        self.assertEqual(gr_mock.call_count, 1)
        self.assertEqual(rs_mock.call_count, 1)
        self.assertEqual(sm_mock.call_count, 1)
        self.assertEqual(sr_mock.call_count, 1)

    @mock.patch('os.rename')
    @mock.patch('os.symlink')
    @mock.patch('dlrn.shell.export_commit_yaml')
    @mock.patch('dlrn.shell.genreports')
    @mock.patch('dlrn.shell.sync_repo')
    def test_successful_build_component(self, rs_mock, gr_mock, ec_mock,
                                        sl_mock, rn_mock):
        built_rpms = ['foo-1.2.3.rpm']
        self.config.use_components = True
        self.commit.component = 'foo'
        status = [self.commit, built_rpms, 'OK', None]
        output = shell.process_build_result(status, self.packages,
                                            self.session, [])
        self.assertEqual(output, 0)
        self.assertEqual(gr_mock.call_count, 1)
        self.assertEqual(rs_mock.call_count, 1)
        self.assertEqual(ec_mock.call_count, 1)
        # 3 additional symlinks due to the current/ directory
        self.assertEqual(sl_mock.call_count, 4)
        self.assertEqual(rn_mock.call_count, 4)


@mock.patch('sh.createrepo_c', create=True)
class TestPostBuild(base.TestCase):
    def setUp(self):
        super(TestPostBuild, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set('DEFAULT', 'datadir', tempfile.mkdtemp())
        config.set('DEFAULT', 'scriptsdir', tempfile.mkdtemp())
        config.set('DEFAULT', 'baseurl', "file://%s" % config.get('DEFAULT',
                                                                  'datadir'))
        self.config = ConfigOptions(config)
        self.commit = db.Commit(dt_commit=123, project_name='foo', type="rpm",
                                commit_hash='1c67b1ab8c6fe273d4e175a14f0df5'
                                            'd3cbbd0edf',
                                repo_dir='/home/dlrn/data/foo',
                                distro_hash='c31d1b18eb5ab5aed6721fc4fad06c9'
                                            'bd242490f',
                                dt_distro=123,
                                distgit_dir='/home/dlrn/data/foo_distro',
                                commit_branch='master', dt_build=1441245153)
        self.db_fd, filepath = tempfile.mkstemp()
        self.session = mocked_session("sqlite:///%s" % filepath)
        self.packages = [{'upstream': 'https://github.com/openstack/foo',
                          'name': 'foo', 'maintainers': 'test@test.com',
                          'master-distgit':
                          'https://github.com/rdo-packages/foo-distgit.git'},
                         {'upstream': 'https://github.com/openstack/test',
                          'name': 'test', 'maintainers': 'test@test.com',
                          'master-distgit':
                          'https://github.com/rdo-packages/test-distgit.git'}]

    def tearDown(self):
        super(TestPostBuild, self).tearDown()
        shutil.rmtree(self.config.datadir)
        shutil.rmtree(self.config.scriptsdir)
        os.close(self.db_fd)

    def test_successful_build(self, sh_mock):
        built_rpms = ['repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.noarch.rpm',
                      'repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.src.rpm']

        status = [self.commit, built_rpms, 'OK', None]
        # Create directory for the CSV file
        yumdir = os.path.join(self.config.datadir, "repos",
                              self.commit.getshardedcommitdir())
        os.makedirs(yumdir)
        output = shell.post_build(status, self.packages,
                                  self.session)

        self.assertTrue(os.path.exists(
                        os.path.join(self.config.datadir,
                                     "repos",
                                     self.commit.getshardedcommitdir(),
                                     "versions.csv")))

        expected = [mock.call(yumdir)]
        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(output, 1)     # 1 non-successfully built package

    def test_successful_build_no_failures(self, sh_mock):
        packages = [{'upstream': 'https://github.com/openstack/foo',
                     'name': 'foo', 'maintainers': 'test@test.com',
                     'master-distgit':
                     'https://github.com/rdo-packages/foo-distgit.git'}]
        built_rpms = ['repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.noarch.rpm',
                      'repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.src.rpm']

        status = [self.commit, built_rpms, 'OK', None]
        # Create directory for the CSV file
        yumdir = os.path.join(self.config.datadir, "repos",
                              self.commit.getshardedcommitdir())
        os.makedirs(yumdir)
        output = shell.post_build(status, packages, self.session)
        expected = [mock.call(yumdir)]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(output, 0)

    def test_successful_build_no_failures_component(self, sh_mock):
        packages = [{'upstream': 'https://github.com/openstack/foo',
                     'name': 'foo', 'maintainers': 'test@test.com',
                     'master-distgit':
                     'https://github.com/rdo-packages/foo-distgit.git'}]
        built_rpms = ['repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.noarch.rpm',
                      'repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.src.rpm']

        self.config.use_components = True
        self.commit.component = 'testcomponent'
        status = [self.commit, built_rpms, 'OK', None]
        # Create directory for the CSV file
        yumdir = os.path.join(self.config.datadir, "repos",
                              self.commit.getshardedcommitdir())
        os.makedirs(yumdir)
        output = shell.post_build(status, packages, self.session)
        expected = [mock.call(yumdir)]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(output, 0)
        repofile = open(os.path.join(yumdir, 'delorean.repo')).read()
        assert 'component-testcomponent' in repofile

    def test_successful_build_no_failures_nosrcrpm(self, sh_mock):
        packages = [{'upstream': 'https://github.com/openstack/foo',
                     'name': 'foo', 'maintainers': 'test@test.com',
                     'master-distgit':
                     'https://github.com/rdo-packages/foo-distgit.git'}]
        built_rpms = ['repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.noarch.rpm',
                      'repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.src.rpm']

        self.config.include_srpm_in_repo = False

        status = [self.commit, built_rpms, 'OK', None]
        # Create directory for the CSV file
        yumdir = os.path.join(self.config.datadir, "repos",
                              self.commit.getshardedcommitdir())
        os.makedirs(yumdir)
        output = shell.post_build(status, packages, self.session)
        expected = [mock.call('-x', '*.src.rpm', yumdir)]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(output, 0)

    def test_successful_build_no_repo(self, sh_mock):
        packages = [{'upstream': 'https://github.com/openstack/foo',
                     'name': 'foo', 'maintainers': 'test@test.com',
                     'master-distgit':
                     'https://github.com/rdo-packages/foo-distgit.git'}]
        built_rpms = ['repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.noarch.rpm',
                      'repos/1c/67/1c67b1ab8c6fe273d4e175a14f0df5d3cbbd0edf'
                      '_c31d1b18/foo-1.2.3.el7.centos.src.rpm']

        status = [self.commit, built_rpms, 'OK', None]
        # Create directory for the CSV file
        yumdir = os.path.join(self.config.datadir, "repos",
                              self.commit.getshardedcommitdir())
        os.makedirs(yumdir)
        output = shell.post_build(status, packages, self.session,
                                  build_repo=False)
        # There will be no createrepo call
        expected = []

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(output, 0)


class TestRecheck(base.TestCase):
    def setUp(self):
        super(TestRecheck, self).setUp()

    def tearDown(self):
        super(TestRecheck, self).tearDown()

    @mock.patch('dlrn.shell.getSession', side_effect=mocked_session_recheck)
    @mock.patch('dlrn.drivers.rdoinfo.RdoInfoDriver.getpackages',
                side_effect=mocked_getpackages)
    def test_basic_recheck(self, gp_mock, db_mock):
        testargs = ['dlrn', '--package-name',
                    'python-pysaml2', '--recheck']
        with mock.patch.object(sys, 'argv', testargs):
            e = self.assertRaises(SystemExit, shell.main)
            # Rechecking python-pysaml2 should fail because the last commit
            # was successfully built
            self.assertEqual(e.code, 1)

    @mock.patch('dlrn.shell.getSession', side_effect=mocked_session_recheck)
    @mock.patch('dlrn.drivers.rdoinfo.RdoInfoDriver.getpackages',
                side_effect=mocked_getpackages)
    def test_force_recheck_withoutprojectsini(self, gp_mock, db_mock):
        testargs = ['dlrn', '--package-name',
                    'python-pysaml2', '--recheck', '--force-recheck']
        with mock.patch.object(sys, 'argv', testargs):
            e = self.assertRaises(SystemExit, shell.main)
            # Rechecking python-pysaml2 should fail because the last commit
            # was successfully built
            self.assertEqual(e.code, 1)

    @mock.patch('dlrn.shell.getSession', side_effect=mocked_session_recheck)
    @mock.patch('dlrn.drivers.rdoinfo.RdoInfoDriver.getpackages',
                side_effect=mocked_getpackages)
    def test_force_recheck_withprojectsini(self, gp_mock, db_mock):
        testargs = ['dlrn', '--package-name',
                    'python-pysaml2', '--recheck', '--force-recheck',
                    '--config-file',
                    './dlrn/tests/samples/projects_force.ini']
        with mock.patch.object(sys, 'argv', testargs):
            e = self.assertRaises(SystemExit, shell.main)
            # It will work this time
            self.assertEqual(e.code, 0)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import shutil
import tempfile

from dlrn.config import ConfigOptions
from dlrn import db
from dlrn import notifications
from dlrn.tests import base
from six.moves import configparser


class TestNotifications(base.TestCase):
    def setUp(self):
        super(TestNotifications, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set('DEFAULT', 'datadir', tempfile.mkdtemp())
        config.set('DEFAULT', 'scriptsdir', tempfile.mkdtemp())
        config.set('DEFAULT', 'baseurl', "file://%s" % config.get('DEFAULT',
                                                                  'datadir'))

        self.config = ConfigOptions(config)
        self.commit = db.Commit(dt_commit=123, project_name='foo',
                                commit_hash='1c67b1ab8c6fe273d4e175a14f0df5'
                                            'd3cbbd0edf',
                                repo_dir='/home/dlrn/data/foo',
                                distro_hash='c31d1b18eb5ab5aed6721fc4fad06c9'
                                            'bd242490f',
                                dt_distro=123,
                                distgit_dir='/home/dlrn/data/foo_distro',
                                commit_branch='master', dt_build=1441245153)
        self.packages = [{'upstream': 'https://github.com/openstack/foo',
                          'name': 'foo', 'maintainers': ['test@test.com'],
                          'master-distgit':
                          'https://github.com/rdo-packages/foo-distgit.git'}]

    def tearDown(self):
        super(TestNotifications, self).tearDown()
        shutil.rmtree(self.config.datadir)
        shutil.rmtree(self.config.scriptsdir)

    @mock.patch('sh.env', create=True)
    def test_submit_review(self, sh_mock):
        notifications.submit_review(self.commit, self.packages, ['FOO=BAR'])
        yumdir = os.path.join(self.config.datadir, 'repos',
                              self.commit.getshardedcommitdir())

        expected = [mock.call(['FOO=BAR', 'GERRIT_URL=https://github.com/'
                                          'openstack/foo/commit/' +
                                          self.commit.commit_hash,
                               'GERRIT_LOG=%s/%s' % (
                                   self.config.baseurl,
                                   self.commit.getshardedcommitdir()
                               ),
                               'GERRIT_MAINTAINERS=test@test.com',
                               'GERRIT_TOPIC=rdo-FTBFS',
                               os.path.join(self.config.scriptsdir,
                                            "submit_review.sh"),
                    self.commit.project_name, yumdir, self.config.datadir,
                    self.config.baseurl,
                    os.path.realpath(self.commit.distgit_dir)],
                    _timeout=300)]

        self.assertEqual(sh_mock.call_count, 1)
        self.assertEqual(sh_mock.call_args_list, expected)

    @mock.patch('smtplib.SMTP', create=True)
    def test_sendnotifymail_nomail(self, smtp_mock):
        notifications.sendnotifymail(self.packages, self.commit)
        # By default there is no smtpserver, so no calls
        self.assertEqual(smtp_mock.call_count, 0)

    @mock.patch('smtplib.SMTP', create=True)
    def test_sendnotifymail(self, smtp_mock):
        self.config.smtpserver = 'foo.example.com'
        notifications.sendnotifymail(self.packages, self.commit)
        self.assertEqual(smtp_mock.call_count, 1)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from dlrn.tests import base
from six.moves import configparser

from dlrn.config import ConfigOptions
from dlrn.config import getConfigOptions
import os


class TestConfigOptions(base.TestCase):
    def setUp(self):
        super(TestConfigOptions, self).setUp()
        self.config = configparser.RawConfigParser()
        self.config.read("projects.ini")

    def test_without_gitrepo_driver(self):
        self.config.remove_section("gitrepo_driver")
        ConfigOptions(self.config)

    def test_with_gitrepo_driver(self):
        self.config.set("DEFAULT", "pkginfo_driver",
                        "dlrn.drivers.gitrepo.GitRepoDriver")
        self.config.set("gitrepo_driver", "skip", "pkg1,pkg2")
        config = ConfigOptions(self.config)
        self.assertEqual(config.skip_dirs, ["pkg1", "pkg2"])

    def test_without_rdoinfo_driver(self):
        self.config.remove_section("rdoinfo_driver")
        ConfigOptions(self.config)

    def test_with_rdoinfo_driver(self):
        self.config.set("DEFAULT", "pkginfo_driver",
                        "dlrn.drivers.rdoinfo.RdoInfoDriver")
        self.config.set(
            "rdoinfo_driver", "repo", "https://test/test.git")
        config = ConfigOptions(self.config)
        self.assertEqual(
            config.rdoinfo_repo, 'https://test/test.git')

    def test_get_config_option(self):
        config = ConfigOptions(self.config)
        self.assertEqual(config, getConfigOptions())

    def test_override(self):
        self.config.set("DEFAULT", "fallback_to_master", "1")
        self.config.set("DEFAULT", "workers", "7")
        config = ConfigOptions(self.config)
        # No changes
        self.assertEqual(config.fallback_to_master, True)
        self.assertEqual(config.workers, 7)

        overrides = ['DEFAULT.fallback_to_master=0',
                     'DEFAULT.workers=3']
        config = ConfigOptions(self.config, overrides=overrides)
        # Overrides in effect
        self.assertEqual(config.fallback_to_master, False)
        self.assertEqual(config.workers, 3)

    def test_override_non_existing(self):
        overrides = ['DEFAULT.non_existing_option=0']
        self.assertRaises(RuntimeError, ConfigOptions,
                          self.config, overrides=overrides)

    def test_dynamic_dirs(self):
        config = ConfigOptions(self.config)
        self.assertEqual(config.scriptsdir, './scripts')
        self.assertEqual(config.configdir, './scripts')
        self.assertEqual(config.templatedir, './dlrn/templates')

    def test_detect_dirs(self):
        self.config = configparser.RawConfigParser()
        self.config.read("samples/projects.ini.detect")
        config = ConfigOptions(self.config)
        self.assertEqual(
            config.datadir,
            os.path.realpath(os.path.join(
                             os.path.dirname(os.path.abspath(__file__)),
                             "../../data")))
        self.assertEqual(
            config.templatedir,
            os.path.realpath(os.path.join(
                             os.path.dirname(os.path.abspath(__file__)),
                             "../templates")))
        self.assertEqual(
            config.scriptsdir,
            os.path.realpath(os.path.join(
                             os.path.dirname(os.path.abspath(__file__)),
                             "../../scripts")))



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import mock
import os
import tempfile

from dlrn import db
from dlrn.tests import base
from dlrn import utils

from sqlalchemy.pool import NullPool


class TestsWithData(base.TestCase):
    def setUp(self):
        super(TestsWithData, self).setUp()
        self.db_fd, filepath = tempfile.mkstemp()
        self.session = db.getSession("sqlite:///%s" % filepath)
        utils.loadYAML(self.session, './dlrn/tests/samples/commits_1.yaml')

    def tearDown(self):
        super(TestsWithData, self).tearDown()
        os.close(self.db_fd)


@mock.patch('sqlalchemy.orm.sessionmaker', autospec=True)
class TestGetSessions(base.TestCase):
    def setUp(self):
        super(TestGetSessions, self).setUp()

    def test_getsession(self, sm_mock):
        db.getSession()
        self.assertEqual(len(sm_mock.call_args_list), 1)

    @mock.patch('sqlalchemy.create_engine')
    def test_getsessions(self, ce_mock, sm_mock):
        db.getSession()
        db.getSession(url="sqlite:///test.db")
        # The 2nd call should now result in a new session
        db.getSession()
        self.assertEqual(len(sm_mock.call_args_list), 3)
        expected = [mock.call('sqlite://', poolclass=NullPool),
                    mock.call('sqlite:///test.db', poolclass=NullPool),
                    mock.call('sqlite://', poolclass=NullPool)]
        self.assertEqual(ce_mock.call_args_list, expected)


class TestGetLastProcessedCommit(TestsWithData):
    def test_noretry(self):
        commit = db.getLastProcessedCommit(self.session, 'python-pysaml2')
        self.assertEqual(commit.dt_build, 1444139517)

    def test_withretry(self):
        # In our sample data the most recent of these has status == RETRY
        commit = \
            db.getLastProcessedCommit(self.session, 'python-tripleoclient')
        self.assertEqual(commit.dt_build, 1444033941)

    def test_newproject(self):
        commit = db.getLastProcessedCommit(self.session, 'python-newproject')
        self.assertEqual(commit, None)


class TestGetLastBuiltCommit(TestsWithData):
    def test_noretry(self):
        commit = \
            db.getLastBuiltCommit(self.session, 'python-pysaml2', 'master')
        self.assertEqual(commit.dt_build, 1444139517)

    def test_withretry(self):
        # In our sample data the most recent of these has status == RETRY
        commit = \
            db.getLastBuiltCommit(self.session, 'python-tripleoclient',
                                  'master')
        self.assertEqual(commit.dt_build, 1444033941)

    def test_newproject(self):
        commit = \
            db.getLastBuiltCommit(self.session, 'python-newproject', 'master')
        self.assertEqual(commit, None)


class TestGetCommits(TestsWithData):
    def test_defaults(self):
        commits = db.getCommits(self.session)
        self.assertEqual(commits.count(), 1)
        self.assertEqual(commits.first().id, 7873)

    def test_no_results(self):
        commits = db.getCommits(self.session, project="dummy")
        self.assertEqual(commits.count(), 0)
        self.assertEqual(commits.first(), None)

    def test_last_success(self):
        commits = db.getCommits(self.session, project="python-tripleoclient",
                                with_status="SUCCESS")
        self.assertEqual(commits.count(), 1)
        self.assertEqual(commits.first().id, 7696)

    def test_last_without_retry(self):
        commits = db.getCommits(self.session, project="python-tripleoclient",
                                without_status="RETRY")
        self.assertEqual(commits.count(), 1)
        self.assertEqual(commits.first().id, 7696)

    def test_last_two(self):
        commits = db.getCommits(self.session, project="python-pysaml2",
                                limit=2)
        self.assertEqual(commits.count(), 2)
        self.assertEqual([c.id for c in commits], [7835, 7834])

    def test_first_failed(self):
        commits = db.getCommits(self.session, project="python-pysaml2",
                                with_status="FAILED", order="asc")
        self.assertEqual(commits.count(), 1)
        self.assertEqual(commits.first().id, 5874)

    def test_first_failed_since(self):
        commits = db.getCommits(self.session, project="python-alembic",
                                with_status="FAILED", order="asc",
                                since="1442487440")
        self.assertEqual(commits.count(), 1)
        self.assertEqual(commits.first().id, 6230)


class TestCommit(TestsWithData):
    def test_commit_compare(self):
        commits = db.getCommits(self.session, project="python-tripleoclient")
        self.assertGreater(commits[0], commits[1])
        self.assertLess(commits[1], commits[0])
        self.assertEqual(commits[0], commits[-1])

    def test_commit_getshardedcommitdir(self):
        commit = db.getLastProcessedCommit(self.session, 'python-pysaml2')
        self.assertIn(commit.commit_hash, commit.getshardedcommitdir())
        commit.distro_hash = None
        self.assertIn(commit.commit_hash, commit.getshardedcommitdir())

    def test_commit_getshardedcommitdir_extended_hash(self):
        commit = db.getLastProcessedCommit(self.session, 'python-pysaml2')
        commit.extended_hash = 'abc'
        directory1 = commit.getshardedcommitdir()
        self.assertEqual(directory1,
                         '3a/93/3a9326f251b9a4162eb0dfa9f1c924ef47c2c55a_'
                         '024e24f0_abc')
        commit.extended_hash = 'abcdef123456'
        directory2 = commit.getshardedcommitdir()
        self.assertEqual(directory2,
                         '3a/93/3a9326f251b9a4162eb0dfa9f1c924ef47c2c55a_'
                         '024e24f0_abcdef12')

    def test_commit_getshardedcommitdir_component(self):
        commit = db.getLastProcessedCommit(self.session, 'python-pysaml2')
        commit.component = 'foo'
        commit.extended_hash = None
        directory1 = commit.getshardedcommitdir()
        self.assertEqual(directory1,
                         'component/foo/3a/93/3a9326f251b9a4162eb0dfa9f1c9'
                         '24ef47c2c55a_024e24f0')


class TestProject(TestsWithData):
    def test_email(self):
        project = self.session.query(db.Project).first()
        self.assertFalse(project.suppress_email())
        project.sent_email()
        self.assertTrue(project.suppress_email())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Set string length

Revision ID: f38ba3389b85
Revises: 1268c799620f
Create Date: 2017-01-12 18:42:25.593667

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f38ba3389b85'
down_revision = '1268c799620f'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("commits") as batch_op:
        batch_op.alter_column('project_name', existing_type=sa.String(),
                              type_=sa.String(256))
        batch_op.alter_column('repo_dir', existing_type=sa.String(),
                              type_=sa.String(1024))
        batch_op.alter_column('distgit_dir', existing_type=sa.String(),
                              type_=sa.String(1024))
        batch_op.alter_column('commit_hash', existing_type=sa.String(),
                              type_=sa.String(64))
        batch_op.alter_column('distro_hash', existing_type=sa.String(),
                              type_=sa.String(64))
        batch_op.alter_column('distgit_dir', existing_type=sa.String(),
                              type_=sa.String(1024))
        batch_op.alter_column('commit_branch', existing_type=sa.String(),
                              type_=sa.String(256))
        batch_op.alter_column('status', existing_type=sa.String(),
                              type_=sa.String(64))
        batch_op.alter_column('rpms', existing_type=sa.String(),
                              type_=sa.Text())
        batch_op.alter_column('notes', existing_type=sa.String(),
                              type_=sa.Text())

    with op.batch_alter_table("projects") as batch_op:
        batch_op.alter_column('project_name', existing_type=sa.String(),
                              type_=sa.String(256))


def downgrade():
    with op.batch_alter_table("commits") as batch_op:
        batch_op.alter_column('project_name', existing_type=sa.String(256),
                              type_=sa.String())
        batch_op.alter_column('repo_dir', existing_type=sa.String(1024),
                              type_=sa.String())
        batch_op.alter_column('distgit_dir', existing_type=sa.String(1024),
                              type_=sa.String())
        batch_op.alter_column('commit_hash', existing_type=sa.String(64),
                              type_=sa.String())
        batch_op.alter_column('distro_hash', existing_type=sa.String(64),
                              type_=sa.String())
        batch_op.alter_column('distgit_dir', existing_type=sa.String(1024),
                              type_=sa.String())
        batch_op.alter_column('commit_branch', existing_type=sa.String(256),
                              type_=sa.String())
        batch_op.alter_column('status', existing_type=sa.String(64),
                              type_=sa.String())
        batch_op.alter_column('rpms', existing_type=sa.Text(),
                              type_=sa.String())
        batch_op.alter_column('notes', existing_type=sa.Text(),
                              type_=sa.String())

    with op.batch_alter_table("projects") as batch_op:
        batch_op.alter_column('project_name', existing_type=sa.String(256),
                              type_=sa.String())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add tables required by DLRN API

Revision ID: 638f980c9169
Revises: f38ba3389b85
Create Date: 2016-11-22 12:43:21.363066

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '638f980c9169'
down_revision = 'f38ba3389b85'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('users',
                    sa.Column('username', sa.String(256), nullable=False),
                    sa.Column('password', sa.String(256), nullable=False),
                    sa.PrimaryKeyConstraint('username'))

    op.create_table('civotes',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('commit_id', sa.Integer(),
                              sa.ForeignKey('commits.id'),
                              nullable=False),
                    sa.Column('ci_name', sa.String(256), nullable=True),
                    sa.Column('ci_url', sa.String(1024), nullable=True),
                    sa.Column('ci_vote', sa.Boolean(), nullable=True),
                    sa.Column('ci_in_progress', sa.Boolean(), nullable=True),
                    sa.Column('timestamp', sa.Integer(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.PrimaryKeyConstraint('id'))


def downgrade():
    op.drop_table('users')
    op.drop_table('civotes')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from dlrn.db import Commit
from dlrn.drivers.pkginfo import PkgInfoDriver
from dlrn.repositories import getdistrobranch
from dlrn.repositories import getsourcebranch
from dlrn.repositories import refreshrepo
from dlrn.utils import run_external_preprocess

import csv
import logging
import os
import re
import sh
import shutil

from distroinfo import info
from distroinfo import query
from six.moves.urllib.request import urlopen


logger = logging.getLogger("dlrn-downstream-driver")


def fail_req_attr_missing(attr_name, package):
    raise Exception(
        "Missing required attribute '%s' "
        "for package: %s (project: %s)" % (
            attr_name, package['project'], package['name']))


def fail_req_config_missing(opt_name):
    raise Exception(
        "Missing required config option '%s' "
        "for dlrn.drivers.downstream driver." % opt_name)


class DownstreamInfoDriver(PkgInfoDriver):
    DRIVER_CONFIG = {
        'downstream_driver': {
            'rdoinfo_repo': {'name': 'repo'},
            'info_files': {'type': 'list'},
            'versions_url': {},
            'downstream_distro_branch': {},
            'downstream_tag': {},
            'downstream_distgit_key': {},
            'use_upstream_spec': {'type': 'boolean'},
            'downstream_spec_replace_list': {'type': 'list'},
            'cache_dir': {},
        }
    }

    def __init__(self, *args, **kwargs):
        super(DownstreamInfoDriver, self).__init__(*args, **kwargs)
        self.distroinfo_path = None

    def getpackages(self, **kwargs):
        """Valid parameters:

        :param local_info_repo: local rdoinfo repo to use instead of fetching
                                the default one using distroinfo.
        :param tags: release tags to use (mitaka, newton, etc).
        """
        local_info_repo = kwargs.get('local_info_repo')
        tags = kwargs.get('tags')

        info_files = self.config_options.info_files
        if not info_files:
            fail_req_config_missing('info_file')

        inforepo = None
        if local_info_repo:
            inforepo = info.DistroInfo(
                info_files=info_files,
                local_info=local_info_repo,
                cache_base_path=self.config_options.cache_dir)
            # NOTE(jpena): in general, info_files will only contain one file,
            # but it supports multiple... In that case, we will have a comma
            # separated list of URLs
            self.distroinfo_path = "%s/%s" % (local_info_repo.rstrip('/'),
                                              info_files[0])
            for extra_file in info_files[1:]:
                self.distroinfo_path += ",%s/%s" % (
                    local_info_repo.rstrip('/'))

        elif self.config_options.rdoinfo_repo:
            inforepo = info.DistroInfo(
                info_files=info_files,
                remote_git_info=self.config_options.rdoinfo_repo,
                cache_base_path=self.config_options.cache_dir)
            rdoinfo_repo = self.config_options.rdoinfo_repo
            self.distroinfo_path = "%s/%s" % (rdoinfo_repo.rstrip('/'),
                                              info_files[0])
            for extra_file in info_files[1:]:
                self.distroinfo_path += ",%s/%s" % (
                    rdoinfo_repo.rstrip('/'), extra_file)

        else:
            fail_req_config_missing('repo')
        pkginfo = inforepo.get_info(apply_tag=tags)

        self.packages = pkginfo["packages"]
        if tags:
            # FIXME allow list of tags?
            self.packages = query.filter_pkgs(self.packages, {'tags': tags})

        if self.config_options.downstream_tag:
            # filter out packages missing parameters with
            # downstream_tag
            downstream_tag = self.config_options.downstream_tag
            self.packages = query.filter_pkgs(
                self.packages,
                {'tags': downstream_tag})
        return self.packages

    def _getversions(self):
        """Fetch 'versions.csv'

        from versions_url config option and return the contained data as
        a dict with package name as a key.
        """
        versions_url = self.config_options.versions_url
        if not versions_url:
            fail_req_config_missing('versions_url')

        # return versions.csv as a dict with package name as a key
        vers = {}
        r = urlopen(versions_url)
        content = [x.decode('utf-8') for x in r.readlines()]
        # first line is headers
        for row in csv.reader(content[1:]):
            vers[row[0]] = row[1:]
        return vers

    def _fetch_reference_commit(self):
        """Fetch the commit.yaml file used for this run

        """
        versions_url = self.config_options.versions_url
        commit_url = re.sub('/versions.csv$', '/commit.yaml', versions_url)
        try:
            r = urlopen(commit_url)
            content = [x.decode('utf-8') for x in r.readlines()]
            return content
        except Exception:
            # We do not want to fail on this
            return None

    def _write_reference_commit(self, directory):
        """Write the commit.yaml file

        Used for this run in directory as reference_commit.yaml

        :param: directory where the file will be stored
        """

        ref_commit_yaml = self._fetch_reference_commit()

        if ref_commit_yaml is None:
            return
        path = os.path.join(directory, 'reference_commit.yaml')
        with open(path, 'w') as fp:
            for line in ref_commit_yaml:
                fp.write(line)

    def _transform_spec(self, directory):
        """Transform based on rules, basically a crude sed implementation

        :param directory: directory to search the spec on
        """
        for f in os.listdir(directory):
            if f.endswith('.spec'):
                specpath = os.path.join(directory, f)
                with open(specpath, "r") as fp:
                    contents = fp.readlines()
                with open(specpath, "w") as fp:
                    for line in contents:
                        for rule in self.config_options.\
                            downstream_spec_replace_list:
                            src = rule.split('/')[0]
                            dst = rule.split('/')[1]
                            line = re.sub(src, dst, line)
                        fp.write(line)

    def getinfo(self, **kwargs):
        project = kwargs.get('project')
        package = kwargs.get('package')
        local = kwargs.get('local')
        dev_mode = kwargs.get('dev_mode')

        datadir = self.config_options.datadir
        repo = package['upstream']
        distgit_attr = self.config_options.downstream_distgit_key or 'distgit'
        distro = package.get(distgit_attr)
        if not distro:
            fail_req_attr_missing(distgit_attr, package)
        distro_dir = self._distgit_clone_dir(package['name'])
        distro_dir_full = self.distgit_dir(package['name'])
        distro_branch = self.config_options.downstream_distro_branch
        if not distro_branch:
            fail_req_config_missing('downstream_distro_branch')
        source_branch = getsourcebranch(package)
        versions = self._getversions()

        ups_distro = package['master-distgit']
        ups_distro_dir = self._upstream_distgit_clone_dir(package['name'])
        ups_distro_dir_full = self.upstream_distgit_dir(package['name'])
        ups_distro_branch = getdistrobranch(package)

        # only process packages present in versions.csv
        if package['name'] not in versions:
            logger.warning('Package %s not present in %s - skipping.' % (
                package['name'], self.config_options.versions_url))
            return []
        version = versions[package['name']]

        dt_distro = 0  # In this driver we do not care about dt_distro

        if dev_mode is False:
            try:
                distro_branch, extended_hash, dt_extended = refreshrepo(
                    distro, distro_dir, distro_branch, local=local,
                    full_path=distro_dir_full)
                # extract distro_hash from versions.csv
                distro_hash = version[3]
                # Also download upstream distgit
                _, _, _ = refreshrepo(
                    ups_distro, ups_distro_dir, distro_hash,
                    local=local, full_path=ups_distro_dir_full)

            except Exception:
                # The error was already logged by refreshrepo, and we want
                # to avoid halting the whole run because this distgit repo
                # failed, so return an empty list
                return []
        else:
            distro_hash = "dev"
            extended_hash = "dev"
            dt_extended = 0
            if not os.path.isdir(distro_dir):
                # We should fail in this case, since we are running
                # in dev mode, so no try/except
                refreshrepo(distro, distro_dir, distro_branch, local=local,
                            full_path=distro_dir_full)
            if not os.path.isdir(ups_distro_dir):
                refreshrepo(ups_distro, ups_distro_dir, ups_distro_branch,
                            local=local, full_path=ups_distro_dir_full)

        # repo is usually a string, but if it contains more then one entry we
        # git clone into a project subdirectory
        repos = [repo]
        if isinstance(repo, list):
            repos = repo
        project_toprocess = []
        for repo in repos:
            repo_dir = os.path.join(datadir, project)
            if len(repos) > 1:
                repo_dir = os.path.join(repo_dir, os.path.split(repo)[1])
            try:
                source_branch, _, _ = refreshrepo(repo, repo_dir,
                                                  source_branch,
                                                  local=local)
            except Exception:
                # The error was already logged by refreshrepo, and the only
                # side-effect is that we are not adding this commit to the
                # list of commits to be processed, so we can ignore it and
                # move on to the next repo
                continue
            if not local:
                # This is the default behavior
                dt = version[5]
                commit_hash = version[1]
            else:
                # When running with --local, we really want to use the local
                # source git, regardless of the upstream versions.csv info
                git = sh.git.bake(_cwd=repo_dir, _tty_out=False)
                lines = git.log("--pretty=format:'%ct %H'",
                                "-1", "--first-parent",
                                "--reverse")
                for line in lines:
                    # There is only one line
                    dt, commit_hash = str(line).strip().strip("'").\
                                        split(" ")[:2]

            if self.config_options.use_components and 'component' in package:
                component = package['component']
            else:
                component = None
            commit = Commit(dt_commit=float(dt), project_name=project,
                            type='rpm',
                            commit_hash=commit_hash, repo_dir=repo_dir,
                            distro_hash=distro_hash, dt_distro=dt_distro,
                            extended_hash=extended_hash,
                            dt_extended=dt_extended,
                            distgit_dir=self.distgit_dir(package['name']),
                            commit_branch=source_branch, component=component)
            project_toprocess.append(commit)

        return project_toprocess

    def preprocess(self, **kwargs):
        package_name = kwargs.get('package_name')
        commit_hash = kwargs.get('commit_hash')
        distgit_dir = self.distgit_dir(package_name)
        ups_distro_dir_full = self.upstream_distgit_dir(package_name)
        distro_dir_full = self.distgit_dir(package_name)
        source_dir = "%s/%s" % (self.config_options.datadir, package_name)

        # In this case, we will copy the upstream distgit into downstream
        # distgit, then transform spec
        if self.config_options.use_upstream_spec:
            if self.config_options.keep_changelog:
                # Save the existing changelog
                changelog = self._save_changelog(distgit_dir)
            # Copy upstream distgit to downstream distgit
            for f in os.listdir(ups_distro_dir_full):
                # skip hidden files
                if not f.startswith('.'):
                    shutil.copy(os.path.join(ups_distro_dir_full, f),
                                distro_dir_full)

            if len(self.config_options.downstream_spec_replace_list) > 0:
                self._transform_spec(distro_dir_full)

            if self.config_options.keep_changelog:
                # Restore the old changelog, instead of replacing it with
                # the upstream one
                self._restore_changelog(distgit_dir, changelog)
        # If we have a jinja2 spec template, run pre-processing
        preprocess_needed = False
        for f in os.listdir(distgit_dir):
            if f.endswith('.spec.j2'):
                # We have a template here, so we have to preprocess
                preprocess_needed = True
                break
        if preprocess_needed:
            logger.info('Pre-processing template at %s' % distgit_dir)
            renderspec = sh.renderspec.bake(_cwd=distgit_dir,
                                            _tty_out=False, _timeout=3600)
            renderspec('--spec-style', 'fedora', '--epoch',
                       '../../epoch/fedora.yaml')

        for custom_preprocess in self.config_options.custom_preprocess:
            if custom_preprocess != '':
                run_external_preprocess(
                    cmdline=custom_preprocess,
                    pkgname=package_name,
                    distgit=distgit_dir,
                    upstream_distgit=ups_distro_dir_full,
                    distroinfo=self.distroinfo_path,
                    source_dir=source_dir,
                    commit_hash=commit_hash)
        return

    def distgit_dir(self, package_name):
        datadir = self.config_options.datadir
        # Find extra directory inside it, if needed
        extra_dir = '/'
        for package in self.packages:
            if package['name'] == package_name:
                if 'distgit-path' in package:
                    extra_dir = package['distgit-path']
                    break
        return os.path.join(datadir, package_name + "_distro",
                            extra_dir.lstrip('/'))

    def _distgit_clone_dir(self, package_name):
        datadir = self.config_options.datadir
        return os.path.join(datadir, package_name + "_distro")

    def upstream_distgit_dir(self, package_name):
        datadir = self.config_options.datadir
        # Find extra directory inside it, if needed
        extra_dir = '/'
        for package in self.packages:
            if package['name'] == package_name:
                if 'distgit-path' in package:
                    extra_dir = package['distgit-path']
                    break
        return os.path.join(datadir, package_name + "_distro_upstream",
                            extra_dir.lstrip('/'))

    def _upstream_distgit_clone_dir(self, package_name):
        datadir = self.config_options.datadir
        return os.path.join(datadir, package_name + "_distro_upstream")

    def _save_changelog(self, distgit_dir):
        changelog = ''
        spec = None
        for f in os.listdir(distgit_dir):
            if f.endswith('.spec'):
                with open(os.path.join(distgit_dir, f)) as fp:
                    spec = fp.read()
                    # This will read everything from a line starting with
                    # %changelog, till the end of file
                    match = re.search(r'^%changelog.*', spec,
                                      flags=(re.MULTILINE | re.DOTALL))
                    if match:
                        changelog = match.group()

        return changelog

    def _restore_changelog(self, distgit_dir, changelog):
        for f in os.listdir(distgit_dir):
            if f.endswith('.spec'):
                with open(os.path.join(distgit_dir, f)) as fp:
                    spec = fp.read()
                # Replace the %changelog part with the saved changelog
                newspec = re.sub(r'^%changelog.*',
                                 changelog,
                                 spec,
                                 flags=(re.MULTILINE | re.DOTALL))
                with open(os.path.join(distgit_dir, f), 'w') as fp:
                    fp.write(newspec)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import shutil
import tempfile

from dlrn.config import ConfigOptions
from dlrn.drivers.downstream import DownstreamInfoDriver
from dlrn.tests import base
from six.moves import configparser


def _mocked_versions(*args, **kwargs):
    fn = './dlrn/tests/samples/versions.csv'
    return open(fn, 'rb')


def _mocked_refreshrepo(*args, **kwargs):
    return 'a', 'b', 'c'


def _mocked_listdir(path):
    return ['openstack-nova.spec']


@mock.patch('dlrn.drivers.downstream.urlopen', side_effect=_mocked_versions)
class TestDriverDownstream(base.TestCase):
    def setUp(self):
        super(TestDriverDownstream, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set("DEFAULT", "pkginfo_driver",
                   "dlrn.drivers.downstream.DownstreamInfoDriver")
        self.temp_dir = tempfile.mkdtemp()
        self.config = ConfigOptions(config)
        self.config.versions_url = \
            'https://trunk.rdoproject.org/centos7-master/current/versions.csv'
        self.config.downstream_distro_branch = 'testbranch'
        self.config.downstream_distgit_base = 'git://git.example.com/rpms'
        self.config.datadir = self.temp_dir
        self.config.use_upstream_spec = False

    def tearDown(self):
        super(TestDriverDownstream, self).tearDown()
        shutil.rmtree(self.temp_dir)

    def test_versions_parse(self, uo_mock):
        driver = DownstreamInfoDriver(cfg_options=self.config)
        versions = driver._getversions()
        nv = versions['openstack-nova']
        assert nv[1] == 'ef6b4f43f467dfad2fd0fe99d9dec3fc93a9ffed', nv[1]
        assert nv[3] == '8fce438abdd12cba33bd9fa4f7d16c098e10094f', nv[3]

    @mock.patch('dlrn.drivers.downstream.refreshrepo',
                side_effect=_mocked_refreshrepo)
    def test_getinfo(self, rr_mock, uo_mock):
        driver = DownstreamInfoDriver(cfg_options=self.config)
        package = {
            'name': 'openstack-nova',
            'project': 'nova',
            'conf': 'rpmfactory-core',
            'upstream': 'git://git.openstack.org/openstack/nova',
            'patches': 'http://review.rdoproject.org/r/p/openstack/nova.git',
            'distgit': 'git://git.example.com/rpms/nova',
            'master-distgit':
                'git://git.example.com/rpms/nova',
            'name': 'openstack-nova',
            'buildsys-tags': [
                'cloud7-openstack-pike-release: openstack-nova-16.1.4-1.el7',
                'cloud7-openstack-pike-testing: openstack-nova-16.1.4-1.el7',
                'cloud7-openstack-queens-release: openstack-nova-17.0.5-1.el7',
                'cloud7-openstack-queens-testing: openstack-nova-17.0.5-1.el7',
            ]
        }
        pkginfo = driver.getinfo(
            package=package,
            project='nova',
            dev_mode=True)

        expected = [mock.call('git://git.example.com/rpms/nova',
                              self.temp_dir + '/openstack-nova_distro',
                              'testbranch',
                              full_path=self.temp_dir + '/openstack-nova_'
                                                        'distro/',
                              local=None),
                    mock.call('git://git.example.com/rpms/nova',
                              self.temp_dir + '/openstack-nova_distro_'
                                              'upstream',
                              'rpm-master',
                              full_path=self.temp_dir + '/openstack-nova'
                                                        '_distro_upstream/',
                              local=None),
                    mock.call('git://git.openstack.org/openstack/nova',
                              self.temp_dir + '/nova', 'master', local=None)]
        if len(rr_mock.call_args_list) == 2:
            # first refreshrepo call is skipped in dev_mode when
            # distro_dir already exists
            expected = expected[1:]
        self.assertEqual(rr_mock.call_args_list, expected)

        pi = pkginfo[0]
        assert pi.commit_hash == 'ef6b4f43f467dfad2fd0fe99d9dec3fc93a9ffed', \
            pi.commit_hash

    @mock.patch('dlrn.drivers.downstream.refreshrepo',
                side_effect=_mocked_refreshrepo)
    def test_getinfo_component(self, rr_mock, uo_mock):
        self.config.use_components = True
        driver = DownstreamInfoDriver(cfg_options=self.config)
        package = {
            'name': 'openstack-nova',
            'project': 'nova',
            'conf': 'rpmfactory-core',
            'upstream': 'git://git.openstack.org/openstack/nova',
            'patches': 'http://review.rdoproject.org/r/p/openstack/nova.git',
            'distgit': 'git://git.example.com/rpms/nova',
            'master-distgit':
                'git://git.example.com/rpms/nova',
            'name': 'openstack-nova',
            'component': 'compute',
            'buildsys-tags': [
                'cloud7-openstack-pike-release: openstack-nova-16.1.4-1.el7',
                'cloud7-openstack-pike-testing: openstack-nova-16.1.4-1.el7',
                'cloud7-openstack-queens-release: openstack-nova-17.0.5-1.el7',
                'cloud7-openstack-queens-testing: openstack-nova-17.0.5-1.el7',
            ]
        }
        pkginfo = driver.getinfo(
            package=package,
            project='nova',
            dev_mode=True)

        pi = pkginfo[0]
        assert pi.component == 'compute'

    @mock.patch('dlrn.drivers.downstream.refreshrepo',
                side_effect=_mocked_refreshrepo)
    def test_getinfo_component_disabled(self, rr_mock, uo_mock):
        self.config.use_components = False
        driver = DownstreamInfoDriver(cfg_options=self.config)
        package = {
            'name': 'openstack-nova',
            'project': 'nova',
            'conf': 'rpmfactory-core',
            'upstream': 'git://git.openstack.org/openstack/nova',
            'patches': 'http://review.rdoproject.org/r/p/openstack/nova.git',
            'distgit': 'git://git.example.com/rpms/nova',
            'master-distgit':
                'git://git.example.com/rpms/nova',
            'name': 'openstack-nova',
            'component': 'compute',
            'buildsys-tags': [
                'cloud7-openstack-pike-release: openstack-nova-16.1.4-1.el7',
                'cloud7-openstack-pike-testing: openstack-nova-16.1.4-1.el7',
                'cloud7-openstack-queens-release: openstack-nova-17.0.5-1.el7',
                'cloud7-openstack-queens-testing: openstack-nova-17.0.5-1.el7',
            ]
        }
        pkginfo = driver.getinfo(
            package=package,
            project='nova',
            dev_mode=True)

        pi = pkginfo[0]
        assert pi.component is None

    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_preprocess_use_upstream_spec(self, ld_mock, uo_mock):
        self.config.use_upstream_spec = True
        self.config.downstream_spec_replace_list =\
            ['^%global with_doc.*/%global with_doc 0']

        # Prepare fake upstream spec
        os.mkdir(os.path.join(self.temp_dir,
                              'openstack-nova_distro_upstream'))
        os.mkdir(os.path.join(self.temp_dir, 'openstack-nova_distro'))
        with open(os.path.join(self.temp_dir,
                               'openstack-nova_distro_upstream',
                               'openstack-nova.spec'), 'w') as fp:
            fp.write("%global with_doc 1\n")
            fp.write("foo")

        driver = DownstreamInfoDriver(cfg_options=self.config)
        driver.preprocess(package_name='openstack-nova')

        # This checks that the spec file got copied over, and modified with
        # downstream_spec_replace_list
        result = open(os.path.join(self.temp_dir, 'openstack-nova_distro',
                      'openstack-nova.spec'), 'r').read()
        expected = '%global with_doc 0\nfoo'
        self.assertEqual(result, expected)

    @mock.patch('os.environ.get', side_effect=['myuser'])
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_custom_preprocess(self, ld_mock, sh_mock, get_mock, uo_mock):
        self.config.custom_preprocess = ['/bin/true']
        driver = DownstreamInfoDriver(cfg_options=self.config)
        driver.preprocess(package_name='foo')

        expected = [mock.call(
            ['DLRN_PACKAGE_NAME=foo',
             'DLRN_DISTGIT=%s/foo_distro/' % self.temp_dir,
             'DLRN_UPSTREAM_DISTGIT=%s/foo_distro_upstream/' % self.temp_dir,
             'DLRN_SOURCEDIR=%s/foo' % self.temp_dir,
             'DLRN_USER=myuser',
             '/bin/true'],
            _cwd='%s/foo_distro/' % self.temp_dir,
            _env={'LANG': 'C'})]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(sh_mock.call_count, 1)

    @mock.patch('os.environ.get', side_effect=['myuser'])
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_custom_preprocess_distroinfo(self, ld_mock, sh_mock, get_mock,
                                          uo_mock):
        self.config.custom_preprocess = ['/bin/true']
        driver = DownstreamInfoDriver(cfg_options=self.config)
        driver.distroinfo_path = '/tmp/test/dsinfo.yml'
        driver.preprocess(package_name='foo')

        expected = [mock.call(
            ['DLRN_PACKAGE_NAME=foo',
             'DLRN_DISTGIT=%s/foo_distro/' % self.temp_dir,
             'DLRN_UPSTREAM_DISTGIT=%s/foo_distro_upstream/' % self.temp_dir,
             'DLRN_DISTROINFO_REPO=/tmp/test/dsinfo.yml',
             'DLRN_SOURCEDIR=%s/foo' % self.temp_dir,
             'DLRN_USER=myuser',
             '/bin/true'],
            _cwd='%s/foo_distro/' % self.temp_dir,
            _env={'LANG': 'C'})]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(sh_mock.call_count, 1)

    @mock.patch('os.environ.get', side_effect=['myuser'])
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    @mock.patch('shutil.copy')
    def test_custom_preprocess_upstream_spec(self, cp_mock, ld_mock, sh_mock,
                                             get_mock, uo_mock):
        self.config.custom_preprocess = ['/bin/true']
        self.config.use_upstream_spec = True
        driver = DownstreamInfoDriver(cfg_options=self.config)
        driver.preprocess(package_name='foo')

        expected = [mock.call(
            ['DLRN_PACKAGE_NAME=foo',
             'DLRN_DISTGIT=%s/foo_distro/' % self.temp_dir,
             'DLRN_UPSTREAM_DISTGIT=%s/foo_distro_upstream/' % self.temp_dir,
             'DLRN_SOURCEDIR=%s/foo' % self.temp_dir,
             'DLRN_USER=myuser',
             '/bin/true'],
            _cwd='%s/foo_distro/' % self.temp_dir,
            _env={'LANG': 'C'})]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(sh_mock.call_count, 1)

    @mock.patch('dlrn.drivers.downstream.DownstreamInfoDriver.'
                '_restore_changelog')
    @mock.patch('dlrn.drivers.downstream.DownstreamInfoDriver._save_changelog')
    @mock.patch('os.environ.get', side_effect=['myuser'])
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    @mock.patch('shutil.copy')
    def test_custom_preprocess_upstream_spec_keep_changelog(self, cp_mock,
                                                            ld_mock, sh_mock,
                                                            get_mock, sc_mock,
                                                            rc_mock, uo_mock):
        self.config.custom_preprocess = ['/bin/true']
        self.config.use_upstream_spec = True
        self.config.keep_changelog = True
        driver = DownstreamInfoDriver(cfg_options=self.config)
        driver.preprocess(package_name='foo')

        expected = [mock.call(
            ['DLRN_PACKAGE_NAME=foo',
             'DLRN_DISTGIT=%s/foo_distro/' % self.temp_dir,
             'DLRN_UPSTREAM_DISTGIT=%s/foo_distro_upstream/' % self.temp_dir,
             'DLRN_SOURCEDIR=%s/foo' % self.temp_dir,
             'DLRN_USER=myuser',
             '/bin/true'],
            _cwd='%s/foo_distro/' % self.temp_dir,
            _env={'LANG': 'C'})]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(sh_mock.call_count, 1)
        self.assertEqual(sc_mock.call_count, 1)
        self.assertEqual(rc_mock.call_count, 1)

    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_custom_preprocess_fail(self, ld_mock, uo_mock):
        self.config.custom_preprocess = ['/bin/nonexistingcommand']
        driver = DownstreamInfoDriver(cfg_options=self.config)
        os.mkdir(os.path.join(self.temp_dir, 'foo_distro'))

        self.assertRaises(RuntimeError, driver.preprocess, package_name='foo')



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import shutil
import tempfile

from dlrn.config import ConfigOptions
from dlrn.drivers.mockdriver import MockBuildDriver
from dlrn.tests import base
from six.moves import configparser


def _mocked_listdir(directory):
    return ['python-pysaml2-3.0-1a.el7.centos.src.rpm']


@mock.patch('sh.restorecon', create=True)
@mock.patch('sh.env', create=True)
@mock.patch('os.listdir', side_effect=_mocked_listdir)
class TestDriverMock(base.TestCase):
    def setUp(self):
        super(TestDriverMock, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set("DEFAULT", "build_driver",
                   "dlrn.drivers.mockdriver.MockBuildDriver")
        self.config = ConfigOptions(config)
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        super(TestDriverMock, self).tearDown()
        shutil.rmtree(self.temp_dir)

    def test_build_package(self, ld_mock, env_mock, rc_mock):
        os.environ['MOCK_CONFIG'] = 'dlrn-1.cfg'
        driver = MockBuildDriver(cfg_options=self.config)
        driver.build_package(output_directory=self.temp_dir)

        datadir = os.path.realpath(self.config.datadir)
        expected = [mock.call('/usr/bin/mock', '-v', '-r',
                              '%s/dlrn-1.cfg' % datadir,
                              '--resultdir', self.temp_dir,
                              '--rebuild',
                              '%s/python-pysaml2-3.0-1a.el7.centos.src.rpm' %
                              self.temp_dir,
                              _err=driver._process_mock_output,
                              _out=driver._process_mock_output,
                              postinstall=True)]
        # 2 sh calls:
        # 1- mock (handled by env_mock)
        # 2- restorecon (handled by rc_mock)
        self.assertEqual(env_mock.call_count, 1)
        self.assertEqual(rc_mock.call_count, 1)
        self.assertEqual(env_mock.call_args_list, expected)

    def test_build_package_noinstall(self, ld_mock, env_mock, rc_mock):
        os.environ['MOCK_CONFIG'] = 'dlrn-1.cfg'
        config_options = self.config
        config_options.install_after_build = False
        driver = MockBuildDriver(cfg_options=config_options)
        driver.build_package(output_directory=self.temp_dir)

        datadir = os.path.realpath(config_options.datadir)
        expected = [mock.call('/usr/bin/mock', '-v', '-r',
                              '%s/dlrn-1.cfg' % datadir,
                              '--resultdir', self.temp_dir,
                              '--rebuild',
                              '%s/python-pysaml2-3.0-1a.el7.centos.src.rpm' %
                              self.temp_dir,
                              _err=driver._process_mock_output,
                              _out=driver._process_mock_output,
                              postinstall=False)]
        # 2 sh calls:
        # 1- mock (handled by env_mock)
        # 2- restorecon (handled by rc_mock)
        self.assertEqual(env_mock.call_count, 1)
        self.assertEqual(rc_mock.call_count, 1)
        self.assertEqual(env_mock.call_args_list, expected)

    def test_build_additional_mock_options(self, ld_mock, env_mock, rc_mock):
        os.environ['MOCK_CONFIG'] = 'dlrn-1.cfg'
        driver = MockBuildDriver(cfg_options=self.config)
        driver.build_package(output_directory=self.temp_dir,
                             additional_mock_opts='-D repo_bootstrap 1')

        datadir = os.path.realpath(self.config.datadir)
        expected = [mock.call('/usr/bin/mock', '-v', '-r',
                              '%s/dlrn-1.cfg' % datadir,
                              '--resultdir', self.temp_dir,
                              '-D repo_bootstrap 1',
                              '--rebuild',
                              '%s/python-pysaml2-3.0-1a.el7.centos.src.rpm' %
                              self.temp_dir,
                              _err=driver._process_mock_output,
                              _out=driver._process_mock_output,
                              postinstall=True)]
        # 2 sh calls:
        # 1- mock (handled by mock_mock)
        # 2- restorecon (handled by rc_mock)
        self.assertEqual(env_mock.call_count, 1)
        self.assertEqual(rc_mock.call_count, 1)
        self.assertEqual(env_mock.call_args_list, expected)

    def test_driver_config(self, ld_mock, env_mock, rc_mock):
        cp = configparser.RawConfigParser()
        cp.read("projects.ini")
        cp.set("DEFAULT", "build_driver",
               "dlrn.drivers.mockdriver.MockBuildDriver")
        # default is True, test override
        cp.set('mockbuild_driver', 'install_after_build', '0')
        config = ConfigOptions(cp)
        self.assertEqual(False, config.install_after_build)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from getpass import getpass
import passlib.hash
from six.moves import configparser
from six.moves import input
import sys

from dlrn.config import ConfigOptions
from dlrn.config import setup_logging
from dlrn.db import closeSession
from dlrn.db import getSession
from dlrn.db import User


def create_user(options, db_connection):
    try:
        session = getSession(db_connection)
        olduser = session.query(User).filter(
            User.username == options.username).first()

        if olduser is None:
            if options.password is None:
                newpass = getpass("Enter password for %s: " %
                                  options.username)
            else:
                newpass = options.password

            password = passlib.hash.sha512_crypt.encrypt(newpass)
            newuser = User(username=options.username,
                           password=password)
            session.add(newuser)
            session.commit()
            closeSession(session)
            print("User %s successfully created" % options.username)
        else:
            print("User %s already exists" % options.username)
            return -1
    except Exception as e:
        print("Failed to create user %s, %s" % (options.username, e))
        return -1
    return 0


def delete_user(options, db_connection):
    session = getSession(db_connection)
    user = session.query(User).filter(
        User.username == options.username).first()

    if user is None:
        print("ERROR: User %s does not exist" % options.username)
        return -1
    else:
        if not options.force:
            print("Are you sure you want to delete user %s? "
                  "If so, type YES to continue." % options.username)
            confirm = input()
            if confirm != "YES":
                print("Action not confirmed, exiting")
                return -1
        session.delete(user)
        session.commit()
        print("User %s deleted" % options.username)
    closeSession(session)
    return 0


def update_user(options, db_connection):
    session = getSession(db_connection)
    password = passlib.hash.sha512_crypt.encrypt(options.password)
    user = session.query(User).filter(
        User.username == options.username).first()

    if user is None:
        print("ERROR: User %s does not exist" % options.username)
        return -1
    else:
        user.password = password
        session.add(user)
        session.commit()
    closeSession(session)
    return 0


command_funcs = {
    'create': create_user,
    'delete': delete_user,
    'update': update_user,
}


def user_manager():
    parser = argparse.ArgumentParser()
    # Some of the non-positional arguments are required, so change the text
    # saying "optional arguments" to just "arguments":
    parser._optionals.title = 'arguments'

    parser.add_argument('--config-file',
                        default='projects.ini',
                        help="Config file. Default: projects.ini")
    parser.add_argument('--debug', action='store_true',
                        help="Print debug logs")

    subparsers = parser.add_subparsers(dest='command',
                                       title='subcommands',
                                       description='available subcommands')
    subparsers.required = True

    # Subcommand create
    parser_create = subparsers.add_parser('create',
                                          help='Create a user')
    parser_create.add_argument('--username', type=str, required=True,
                               help='User name')
    parser_create.add_argument('--password', type=str, help='Password')

    # Subcommand delete
    parser_delete = subparsers.add_parser('delete',
                                          help='Delete a user')
    parser_delete.add_argument('--username', type=str, required=True,
                               help='User name')
    parser_delete.add_argument('--force', dest='force',
                               action='store_true',
                               help='Do not request a confirmation')

    # Subcommand update
    parser_update = subparsers.add_parser('update',
                                          help='Update a user')
    parser_update.add_argument('--username', type=str, required=True,
                               help='User name')
    parser_update.add_argument('--password', type=str, required=True,
                               help='New password')

    options = parser.parse_args(sys.argv[1:])

    setup_logging(options.debug)

    cp = configparser.RawConfigParser()
    cp.read(options.config_file)
    config_options = ConfigOptions(cp)

    return command_funcs[options.command](options,
                                          config_options.database_connection)



#
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''Basic rpm spec file parsing to be able to get the package names and
the dependencies.
'''

from __future__ import print_function
import os
import re
import sys

name_regexp = re.compile(r'^Name:\s*(.+)\s*$')
package_regexp = re.compile(r'^(Provides:|%package)\s+(-n)?\s*(.+)\s*$')
define_regexp = re.compile(r'[^#]*%(?:define|global)\s+(.+)\s+(.+)\s*$')
build_requires_regexp = re.compile(r'^(?:Build)?Requires(?:\([^\)]+\))?:'
                                   r'\s*(.+)\s*$')


class RpmSpecFile(object):
    def __init__(self, content):
        self._defines = {}
        self._packages = []
        self._build_requires = []
        self.name = '<none>'
        self._provided = {}
        for line in content.split('\n'):
            # lookup macros
            res = define_regexp.search(line)
            if res:
                self._defines[res.group(1)] = res.group(2)
            else:
                # lookup Name:
                res = name_regexp.search(line)
                if res:
                    if self.name == '<none>':
                        self.name = self._expand_defines(res.group(1))
                        self._packages.append(self.name)
                        self._defines['name'] = self.name
                else:
                    # lookup package
                    res = package_regexp.search(line)
                    if res:
                        pkg = re.split(r'\s+|[><=]',
                                       self._expand_defines(res.group(3)))[0]
                        if res.group(2) or res.group(1) == 'Provides:':
                            pkg_name = pkg
                        else:
                            pkg_name = self.name + '-' + pkg
                        self._packages.append(pkg_name)
                        self._provided[pkg_name] = self.name
                    else:
                        # lookup BuildRequires:
                        res = build_requires_regexp.search(line)
                        if res:
                            # split requires on the same lines and
                            # remove >=, <= or = clauses
                            self._build_requires.extend(
                                [re.split(r'\s+|[><=]',
                                          self._expand_defines(req))[0]
                                 for req in re.split(r'\s*,\s*',
                                                     res.group(1))])
        # remove dups
        self._build_requires = list(set(self._build_requires))

    def _expand_defines(self, content):
        lookup_start = content.find('%{')
        lookup_end = content.find('}', lookup_start)
        if (content[lookup_start:lookup_start + 2] ==
           '%{' and content[lookup_end] == '}'):
            return self._expand_defines(
                content[:lookup_start] +
                self._defines.get(content[lookup_start + 2:lookup_end], '') +
                content[lookup_end + 1:])
        return content

    def packages(self):
        return self._packages

    def build_requires(self):
        return self._build_requires

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name


class RpmSpecCollection(object):
    def __init__(self, initial_list=None, debug=False):
        if initial_list:
            self.specs = initial_list
        else:
            self.specs = []
        self.debug = debug

    def compute_order(self):
        # sub-package to package associative array
        self.pkg = {}
        # list to return the computed order
        self.order = []
        # keep status of the progress by
        # package: 0 not processed, 1
        # processing dependencies, 2 already
        # processed
        self.color = {}

        for spec in sorted(self.specs):
            self.pkg[spec.name] = spec
            self.color[spec.name] = 0
            for pkg_name in spec.packages():
                self.pkg[pkg_name] = spec
        for spec in self.specs:
            if self.color[spec.name] == 0:
                self._visit(spec)
        return self.order

    def _visit(self, spec):
        if self.color[spec.name] == 1:
            sys.stderr.write('cycle detected on %s\n' %
                             ', '.join([k for k in self.color
                                        if self.color[k] == 1]))
        elif self.color[spec.name] == 2:
            return
        else:
            self.color[spec.name] = 1
            for breq in spec.build_requires():
                if breq in spec.packages():
                    continue
                if breq in self.pkg:
                    self._visit(self.pkg[breq])
        self.color[spec.name] = 2
        self.order.append(spec.name)


def _main():
    if len(sys.argv) > 2:
        # output the graph in a graphviz compliant format if -g is passed
        if sys.argv[1] == '-g':
            specs = RpmSpecCollection([RpmSpecFile(open(arg).read(-1))
                                       for arg in sys.argv[2:]],
                                      debug=os.getenv('DEBUG'))
            specnames = specs.compute_order()
            print('digraph G {')
            for spec in specs.specs:
                if spec.name not in specnames:
                    continue
                for breq in spec.build_requires():
                    if breq not in spec.packages() and breq in specs.pkg:
                        print('  "%s" -> "%s";' %
                              (spec.name, specs.pkg[breq].name))
            print('}')
        else:
            specs = RpmSpecCollection([RpmSpecFile(open(arg).read(-1))
                                       for arg in sys.argv[1:]],
                                      debug=os.getenv('DEBUG'))
            print('Build order:')
            print('\n'.join(specs.compute_order()))
    else:
        spec = RpmSpecFile(open(sys.argv[1]).read(-1))
        print('Packages:', ', '.join(spec.packages()))
        print('BuildRequires:', ', '.join(spec.build_requires()))


if __name__ == "__main__":
    _main()

# rpmspecfile.py ends here



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import sys

from tempfile import mkstemp

from dlrn.db import getSession
from dlrn.utils import loadYAML
from dlrn.utils import saveYAML


def migrate_db(source_string, dest_string):
    osfd, tmpfilename = mkstemp()
    session = getSession(source_string)
    saveYAML(session, tmpfilename)
    session.close()
    session2 = getSession(dest_string)
    loadYAML(session2, tmpfilename)
    os.remove(tmpfilename)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--source',
                        help="SQLAlchemy connection string for the source"
                             "database.",
                        required=True)
    parser.add_argument('--dest',
                        help="SQLAlchemy connection string for the"
                             "destination database.",
                        required=True)

    options, args = parser.parse_known_args(sys.argv[1:])

    migrate_db(options.source, options.dest)


if __name__ == '__main__':
    main()
    exit(0)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import shutil
import tempfile

from dlrn.config import ConfigOptions
from dlrn.drivers.rdoinfo import RdoInfoDriver
from dlrn.tests import base
from six.moves import configparser


def _mocked_listdir(path):
    return ['foo.spec']


class TestDriverRdoInfo(base.TestCase):
    def setUp(self):
        super(TestDriverRdoInfo, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set("DEFAULT", "pkginfo_driver",
                   "dlrn.drivers.rdoinfo.RdoInfoDriver")
        self.temp_dir = tempfile.mkdtemp()
        self.config = ConfigOptions(config)
        self.config.datadir = self.temp_dir

    def tearDown(self):
        super(TestDriverRdoInfo, self).tearDown()
        shutil.rmtree(self.temp_dir)

    @mock.patch('distroinfo.info.DistroInfo')
    def test_getpackages_default(self, di_mock):
        driver = RdoInfoDriver(cfg_options=self.config)
        driver.getpackages()
        rdoinfo_repo = ('https://raw.githubusercontent.com/'
                        'redhat-openstack/rdoinfo/master/')

        expected = [mock.call(info_files=['rdo.yml'],
                              remote_info=rdoinfo_repo,
                              cache_base_path=None)]

        self.assertEqual(di_mock.call_args_list, expected)

    @mock.patch('distroinfo.info.DistroInfo')
    def test_getpackages_config(self, di_mock):
        self.config.rdoinfo_file = ['foo.yml']
        self.config.rdoinfo_repo = 'http:/github.com/foo'
        driver = RdoInfoDriver(cfg_options=self.config)
        driver.getpackages()

        expected = [mock.call(info_files=['foo.yml'],
                              remote_git_info='http:/github.com/foo',
                              cache_base_path=None)]

        self.assertEqual(di_mock.call_args_list, expected)

    @mock.patch('distroinfo.info.DistroInfo')
    def test_getpackages_local_info_repo(self, di_mock):
        self.config.rdoinfo_file = ['foo.yml']
        driver = RdoInfoDriver(cfg_options=self.config)
        driver.getpackages(local_info_repo='/tmp/bar')

        expected = [mock.call(info_files=['foo.yml'],
                              local_info='/tmp/bar',
                              cache_base_path=None)]

        self.assertEqual(di_mock.call_args_list, expected)

    @mock.patch('os.environ.get', side_effect=['myuser'])
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_custom_preprocess(self, ld_mock, sh_mock, get_mock):
        self.config.custom_preprocess = ['/bin/true']
        driver = RdoInfoDriver(cfg_options=self.config)
        driver.preprocess(package_name='foo')

        expected = [mock.call(['DLRN_PACKAGE_NAME=foo',
                               'DLRN_DISTGIT=%s/foo_distro/' % self.temp_dir,
                               'DLRN_SOURCEDIR=%s/foo' % self.temp_dir,
                               'DLRN_USER=myuser',
                               '/bin/true'],
                              _cwd='%s/foo_distro/' % self.temp_dir,
                              _env={'LANG': 'C'})]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(sh_mock.call_count, 1)

    @mock.patch('os.environ.get', side_effect=['myuser'])
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_custom_preprocess_source_commit(self, ld_mock, sh_mock,
                                             get_mock):
        self.config.custom_preprocess = ['/bin/true']
        driver = RdoInfoDriver(cfg_options=self.config)
        driver.preprocess(package_name='foo', commit_hash='abc123456')

        expected = [mock.call(['DLRN_PACKAGE_NAME=foo',
                               'DLRN_DISTGIT=%s/foo_distro/' % self.temp_dir,
                               'DLRN_SOURCEDIR=%s/foo' % self.temp_dir,
                               'DLRN_SOURCE_COMMIT=abc123456',
                               'DLRN_USER=myuser',
                               '/bin/true'],
                              _cwd='%s/foo_distro/' % self.temp_dir,
                              _env={'LANG': 'C'})]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(sh_mock.call_count, 1)

    @mock.patch('os.environ.get', side_effect=['myuser'])
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_custom_preprocess_rdoinfo_repo(self, ld_mock, sh_mock, get_mock):
        self.config.custom_preprocess = ['/bin/true']
        driver = RdoInfoDriver(cfg_options=self.config)
        driver.distroinfo_path = '/tmp/test/rdo.yml'
        driver.preprocess(package_name='foo')

        expected = [mock.call(['DLRN_PACKAGE_NAME=foo',
                               'DLRN_DISTGIT=%s/foo_distro/' % self.temp_dir,
                               'DLRN_DISTROINFO_REPO=/tmp/test/rdo.yml',
                               'DLRN_SOURCEDIR=%s/foo' % self.temp_dir,
                               'DLRN_USER=myuser',
                               '/bin/true'],
                              _cwd='%s/foo_distro/' % self.temp_dir,
                              _env={'LANG': 'C'})]

        self.assertEqual(sh_mock.call_args_list, expected)
        self.assertEqual(sh_mock.call_count, 1)

    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_custom_preprocess_multiple_commands(self, ld_mock, sh_mock):
        self.config.custom_preprocess = ['/bin/true', '/bin/true',
                                         '/bin/true']
        driver = RdoInfoDriver(cfg_options=self.config)
        driver.preprocess(package_name='foo')
        self.assertEqual(sh_mock.call_count, 3)

    @mock.patch('os.listdir', side_effect=_mocked_listdir)
    def test_custom_preprocess_fail(self, ld_mock):
        self.config.custom_preprocess = ['/bin/nonexistingcommand']
        driver = RdoInfoDriver(cfg_options=self.config)
        os.mkdir(os.path.join(self.temp_dir, 'foo_distro'))

        self.assertRaises(RuntimeError, driver.preprocess, package_name='foo')



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from mock import call
from mock import MagicMock

import hashlib
import os
import shutil
import tempfile

from dlrn import db
from dlrn.tests import base
from dlrn.tests.test_db import TestsWithData
from dlrn import utils


class Testdumpshas2file(TestsWithData):
    def test_noretry(self):
        # SUCCESS
        commit = db.getLastProcessedCommit(self.session, 'python-pysaml2')
        # FAILED
        commit2 = db.getLastProcessedCommit(self.session, 'python-alembic')
        # SUCCESS, RETRY (should be ignored)
        commit3 = \
            db.getLastProcessedCommit(self.session, 'python-tripleoclient')

        mock_fp = MagicMock()
        utils.dumpshas2file(mock_fp, commit, "a", "b", commit.status, 0,
                            None,
                            ['python-saml2-1.0-1.el7.src.rpm'])
        utils.dumpshas2file(mock_fp, commit2, "a", "b", commit2.status, 1,
                            'common',
                            ['python-alembic-1.0-2.el7.src.rpm'])
        utils.dumpshas2file(mock_fp, commit3, "a", "b", commit3.status, 2,
                            'common',
                            ['file1-1.2-3.el7.noarch.rpm',
                             'file2-1.2-3.el7.src.rpm'])
        expected = [
            call.write(u'python-pysaml2,a,3a9326f251b9a4162eb0dfa9f1c924ef47c'
                       '2c55a,b,024e24f0cf4366c2290c22f24e42de714d1addd1'
                       ',SUCCESS,0,None,cafecafe,python-saml2-1.0-1.el7\n'),
            call.write(u'python-alembic,a,459549c9ab7fef91b2dc8986bc0643bb2f6'
                       'ec0c8,b,885e80778edb6cbb8ee4d8909623be8062369a04'
                       ',FAILED,1,common,None,python-alembic-1.0-2.el7\n'),
            call.write(u'python-tripleoclient,a,1da7b10e55abf8c518e8f61ee7966'
                       '188f0405f59,b,0b1ce934e5b2e7d45a448f6555d24036f9aeca51'
                       ',SUCCESS,2,common,None,file2-1.2-3.el7\n')
        ]
        self.assertEqual(mock_fp.mock_calls, expected)


class TestIsKnownError(base.TestCase):
    def setUp(self):
        super(TestIsKnownError, self).setUp()
        self.logfile = tempfile.mkstemp()[1]

    def tearDown(self):
        super(TestIsKnownError, self).tearDown()
        os.unlink(self.logfile)

    def test_isknownerror(self):
        self.assertFalse(utils.isknownerror("/unkownfile"),
                         msg="isknownerror didn't succeed on unknown file")

        with open(self.logfile, "w") as fp:
            fp.write("Error: Nothing to do")
        self.assertTrue(utils.isknownerror(self.logfile),
                        msg="isknownerror didn't find an error")

        with open(self.logfile, "w") as fp:
            fp.write("Success")
        self.assertFalse(utils.isknownerror(self.logfile),
                         msg="isknownerror found unknown error")


class TestAggregateRepo(base.TestCase):
    def setUp(self):
        super(base.TestCase, self).setUp()
        self.db_fd, filepath = tempfile.mkstemp()
        self.session = db.getSession("sqlite:///%s" % filepath)
        utils.loadYAML(self.session,
                       './dlrn/tests/samples/commits_components.yaml')
        self.datadir = tempfile.mkdtemp()
        self.repodir = os.path.join(self.datadir,
                                    'repos/component/tripleo/test1')
        os.makedirs(self.repodir)
        with open(os.path.join(self.repodir, "delorean.repo"), 'w') as fp:
            fp.write("TESTING ONE TWO THREE")

    def tearDown(self):
        super(base.TestCase, self).tearDown()
        shutil.rmtree(self.datadir)
        os.close(self.db_fd)

    def test_aggregate_repo_files(self):
        result = utils.aggregate_repo_files('test1', self.datadir,
                                            self.session, 'delorean')
        expected_file = os.path.join(self.datadir, 'repos', 'test1',
                                     'delorean.repo')
        assert os.path.exists(expected_file)
        with open(expected_file, 'r') as fp:
            contents = fp.read()
        assert contents == 'TESTING ONE TWO THREE\n'
        assert result == hashlib.md5(b'TESTING ONE TWO THREE\n').hexdigest()

    def test_aggregate_repo_files_hashed_dir(self):
        utils.aggregate_repo_files('test1', self.datadir, self.session,
                                   'delorean', hashed_dir=True)
        expected_file = os.path.join(self.datadir, 'repos', 'test1',
                                     'delorean.repo')
        assert os.path.exists(expected_file)
        assert os.path.islink(expected_file)
        with open(expected_file, 'r') as fp:
            contents = fp.read()
        assert contents == 'TESTING ONE TWO THREE\n'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime
from datetime import timedelta
from distutils.util import strtobool
from functools import wraps

from dlrn.api import app
from dlrn.api.utils import AggDetail
from dlrn.api.utils import auth
from dlrn.api.utils import InvalidUsage
from dlrn.api.utils import RepoDetail

from dlrn.config import ConfigOptions
from dlrn.db import CIVote
from dlrn.db import CIVote_Aggregate
from dlrn.db import closeSession
from dlrn.db import Commit
from dlrn.db import getCommits
from dlrn.db import getSession
from dlrn.db import Promotion
from dlrn.purge import FLAG_PURGED
from dlrn.remote import import_commit
from dlrn.utils import aggregate_repo_files
from dlrn.utils import find_in_artifacts

from flask import jsonify
from flask import render_template
from flask import request

import calendar
import os
import re
from six.moves import configparser
from sqlalchemy import desc
import time


pagination_limit = 100
max_limit = 100


def _get_config_options(config_file):
    cp = configparser.RawConfigParser()
    cp.read(config_file)
    return ConfigOptions(cp)


def _repo_hash(commit):
    return "%s_%s" % (commit.commit_hash, commit.distro_hash[:8])


def _json_media_type(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if request.headers.get('Content-Type') != 'application/json':
            raise InvalidUsage('Unsupported Media Type, use JSON',
                               status_code=415)
        return f(*args, **kwargs)
    return decorated_function


def _get_commit(session, commit_hash, distro_hash):
    commit = session.query(Commit).filter(
        Commit.status == 'SUCCESS',
        Commit.commit_hash == commit_hash,
        Commit.distro_hash == distro_hash).order_by(desc(Commit.id)).first()
    return commit


def _rollback_batch_promotion(rollback_list):
    # We want to roll everything back in reverse order
    # We will try our best, but there is never a 100% guarantee
    for item in reversed(rollback_list):
        target_link = item['target_link']
        previous_link = item['previous_link']
        # Remove new link
        try:
            os.remove(target_link)
        except Exception:
            pass    # yes, ignore errors
        # Re-establish old link, if it existed
        if previous_link:
            try:
                os.symlink(previous_link, target_link)
            except Exception:
                pass    # yes, ignore errors


@app.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


def getVote(session, timestamp, success=None, job_id=None, component=None,
            fallback=True):
    votes = session.query(CIVote)
    votes = votes.filter(CIVote.timestamp > timestamp)
    # Initially we want to get any tested repo, excluding consistent repos
    votes = votes.filter(CIVote.ci_name != 'consistent')
    if success is not None:
        votes = votes.filter(CIVote.ci_vote == int(success))
    if job_id is not None:
        votes = votes.filter(CIVote.ci_name == job_id)
    if component is not None:
        votes = votes.filter(CIVote.component == component)
    vote = votes.order_by(desc(CIVote.timestamp)).first()

    if vote is None and not fallback:
        # This is the sequential use case. We do not want to find any vote
        # for a different CI
        raise InvalidUsage('No vote found', status_code=404)

    if vote is None and job_id is not None:
        # Second chance: no votes found for job_id. Let's find any real CI
        # vote, other than 'consistent'
        votes = session.query(CIVote).filter(CIVote.timestamp > timestamp)
        if success is not None:
            votes = votes.filter(CIVote.ci_vote == success)
        votes.filter(CIVote.ci_name != 'consistent')
        vote = votes.order_by(desc(CIVote.timestamp)).first()

    if vote is None:
        # No votes found, let's try to find one for consistent
        votes = session.query(CIVote).filter(CIVote.timestamp > timestamp)
        if success is not None:
            votes = votes.filter(CIVote.ci_vote == success)
        if component is not None:
            votes = votes.filter(CIVote.component == component)
        votes.filter(CIVote.ci_name == 'consistent')
        vote = votes.order_by(desc(CIVote.timestamp)).first()

    if vote is None:
        # No Votes found at all
        raise InvalidUsage('No vote found', status_code=404)

    return vote


@app.route('/api/repo_status', methods=['GET'])
def repo_status():
    # commit_hash: commit hash
    # distro_hash: distro hash
    # success(optional): only report successful/unsuccessful votes

    commit_hash = request.args.get('commit_hash', None)
    distro_hash = request.args.get('distro_hash', None)
    success = request.args.get('success', None)

    if request.headers.get('Content-Type') == 'application/json':
        # This is the old, deprecated method of in-body parameters
        # We will keep it for backwards compatibility
        if commit_hash is None:
            commit_hash = request.json.get('commit_hash', None)
        if distro_hash is None:
            distro_hash = request.json.get('distro_hash', None)
        if success is None:
            success = request.json.get('success', None)

    if (commit_hash is None or distro_hash is None):
        raise InvalidUsage('Missing parameters', status_code=400)

    if success is not None:
        success = bool(strtobool(success))

    # Find the commit id for commit_hash/distro_hash
    session = getSession(app.config['DB_PATH'])
    commit = _get_commit(session, commit_hash, distro_hash)

    if commit is None:
        raise InvalidUsage('commit_hash+distro_hash combination not found',
                           status_code=404)
    commit_id = commit.id

    # Now find every vote for this commit_hash/distro_hash combination
    votes = session.query(CIVote).filter(CIVote.commit_id == commit_id)
    if success is not None:
        votes = votes.filter(CIVote.ci_vote == int(success))

    # And format the output
    data = []
    for vote in votes:
        d = {'timestamp': vote.timestamp,
             'commit_hash': commit_hash,
             'distro_hash': distro_hash,
             'job_id': vote.ci_name,
             'success': bool(vote.ci_vote),
             'in_progress': vote.ci_in_progress,
             'url': vote.ci_url,
             'notes': vote.notes,
             'user': vote.user,
             'component': vote.component}
        data.append(d)
    closeSession(session)
    return jsonify(data)


@app.route('/api/agg_status', methods=['GET'])
def agg_status():
    # aggregate_hash: aggregate hash
    # success(optional): only report successful/unsuccessful votes
    agg_hash = request.args.get('aggregate_hash', None)
    success = request.args.get('success', None)

    if request.headers.get('Content-Type') == 'application/json':
        # This is the old, deprecated method of in-body parameters
        # We will keep it for backwards compatibility
        if agg_hash is None:
            agg_hash = request.json.get('aggregate_hash', None)
        if success is None:
            success = request.json.get('success', None)

    if agg_hash is None:
        raise InvalidUsage('Missing parameters', status_code=400)

    if success is not None:
        success = bool(strtobool(success))

    # Find the aggregates
    session = getSession(app.config['DB_PATH'])
    votes = session.query(CIVote_Aggregate)
    votes = votes.filter(CIVote_Aggregate.ref_hash == agg_hash)
    if success is not None:
        votes = votes.filter(CIVote_Aggregate.ci_vote == int(success))

    # And format the output
    data = []
    for vote in votes:
        d = {'timestamp': vote.timestamp,
             'aggregate_hash': agg_hash,
             'job_id': vote.ci_name,
             'success': bool(vote.ci_vote),
             'in_progress': vote.ci_in_progress,
             'url': vote.ci_url,
             'notes': vote.notes,
             'user': vote.user}
        data.append(d)
    closeSession(session)
    return jsonify(data)


@app.route('/api/last_tested_repo', methods=['GET'])
def last_tested_repo_GET():
    # max_age: Maximum age in hours, used as base for the search
    # success(optional): find repos with a successful/unsuccessful vote
    # job_id(optional); name of the CI that sent the vote
    # sequential_mode(optional): if set to true, change the search algorithm
    #                            to only use previous_job_id as CI name to
    #                            search for. Defaults to false
    # previous_job_id(optional): CI name to search for, if sequential_mode is
    #                            True
    # component(optional): only get votes for this component

    max_age = request.args.get('max_age', None)
    job_id = request.args.get('job_id', None)
    success = request.args.get('success', None)
    sequential_mode = request.args.get('sequential_mode', None)
    previous_job_id = request.args.get('previous_job_id', None)
    component = request.args.get('component', None)

    if request.headers.get('Content-Type') == 'application/json':
        # This is the old, deprecated method of in-body parameters
        # We will keep it for backwards compatibility
        if max_age is None:
            max_age = request.json.get('max_age', None)
        if job_id is None:
            job_id = request.json.get('job_id', None)
        if success is None:
            success = request.json.get('success', None)
        if sequential_mode is None:
            sequential_mode = request.json.get('sequential_mode', None)
        if previous_job_id is None:
            previous_job_id = request.json.get('previous_job_id', None)
        if component is None:
            component = request.json.get('component', None)

    if success is not None:
        success = bool(strtobool(success))

    if sequential_mode is not None:
        sequential_mode = bool(strtobool(sequential_mode))

    if sequential_mode and previous_job_id is None:
        raise InvalidUsage('Missing parameter previous_job_id',
                           status_code=400)

    if max_age is None:
        raise InvalidUsage('Missing parameters', status_code=400)

    # Calculate timestamp as now - max_age
    if int(max_age) == 0:
        timestamp = 0
    else:
        oldest_time = datetime.now() - timedelta(hours=int(max_age))
        timestamp = time.mktime(oldest_time.timetuple())

    session = getSession(app.config['DB_PATH'])
    try:
        if sequential_mode:
            # CI pipeline case
            vote = getVote(session, timestamp, success, previous_job_id,
                           component=component, fallback=False)
        else:
            # Normal case
            vote = getVote(session, timestamp, success, job_id,
                           component=component)
    except Exception as e:
        raise e

    commit = session.query(Commit).filter(
        Commit.status == 'SUCCESS',
        Commit.id == vote.commit_id).first()

    result = {'commit_hash': commit.commit_hash,
              'distro_hash': commit.distro_hash,
              'timestamp': vote.timestamp,
              'job_id': vote.ci_name,
              'success': vote.ci_vote,
              'in_progress': vote.ci_in_progress,
              'user': vote.user,
              'component': vote.component}
    closeSession(session)
    return jsonify(result), 200


@app.route('/api/promotions', methods=['GET'])
def promotions_GET():
    # commit_hash(optional): commit hash
    # distro_hash(optional): distro hash
    # aggregate_hash(optional): aggregate hash
    # promote_name(optional): only report promotions for promote_name
    # offset(optional): skip the first X promotions (only 100 are shown
    #                   per query)
    # limit(optional): maximum number of entries to return
    # component(optional): only report promotions for this component
    commit_hash = request.args.get('commit_hash', None)
    distro_hash = request.args.get('distro_hash', None)
    agg_hash = request.args.get('aggregate_hash', None)
    promote_name = request.args.get('promote_name', None)
    offset = int(request.args.get('offset', 0))
    limit = int(request.args.get('limit', 100))
    component = request.args.get('component', None)

    if request.headers.get('Content-Type') == 'application/json':
        # This is the old, deprecated method of in-body parameters
        # We will keep it for backwards compatibility
        if commit_hash is None:
            commit_hash = request.json.get('commit_hash', None)
        if distro_hash is None:
            distro_hash = request.json.get('distro_hash', None)
        if agg_hash is None:
            agg_hash = request.json.get('aggregate_hash', None)
        if promote_name is None:
            promote_name = request.json.get('promote_name', None)
        if offset == 0:
            offset = int(request.json.get('offset', 0))
        if limit == 100:
            limit = int(request.json.get('limit', 100))
        if component is None:
            component = request.json.get('component', None)

    config_options = _get_config_options(app.config['CONFIG_FILE'])

    # Make sure we do not exceed
    if limit > max_limit:
        limit = max_limit

    if ((commit_hash and not distro_hash) or
            (distro_hash and not commit_hash)):

        raise InvalidUsage('Both commit_hash and distro_hash must be '
                           'specified if any of them is.',
                           status_code=400)

    # Find the commit id for commit_hash/distro_hash
    session = getSession(app.config['DB_PATH'])
    if commit_hash and distro_hash:
        commit = _get_commit(session, commit_hash, distro_hash)
        if commit is None:
            raise InvalidUsage('commit_hash+distro_hash combination not found',
                               status_code=404)
        commit_id = commit.id
    else:
        commit_id = None

    # Now find the promotions, and filter if necessary
    promotions = session.query(Promotion)
    if commit_id is not None:
        promotions = promotions.filter(Promotion.commit_id == commit_id)
    if promote_name is not None:
        promotions = promotions.filter(
            Promotion.promotion_name == promote_name)
    if agg_hash is not None:
        promotions = promotions.filter(Promotion.aggregate_hash == agg_hash)
    if component is not None:
        promotions = promotions.filter(Promotion.component == component)

    promotions = promotions.order_by(desc(Promotion.id)).limit(limit).\
        offset(offset)

    # And format the output
    data = []
    for promotion in promotions:
        commit = getCommits(session, limit=0).filter(
            Commit.id == promotion.commit_id).first()

        repo_hash = _repo_hash(commit)
        repo_url = "%s/%s" % (config_options.baseurl,
                              commit.getshardedcommitdir())

        d = {'timestamp': promotion.timestamp,
             'commit_hash': commit.commit_hash,
             'distro_hash': commit.distro_hash,
             'aggregate_hash': promotion.aggregate_hash,
             'repo_hash': repo_hash,
             'repo_url': repo_url,
             'promote_name': promotion.promotion_name,
             'component': promotion.component,
             'user': promotion.user}
        data.append(d)
    closeSession(session)
    return jsonify(data)


@app.route('/api/metrics/builds', methods=['GET'])
def get_metrics():
    # start_date: start date for period, in YYYY-mm-dd format (UTC)
    # end_date: end date for period, in YYYY-mm-dd format (UTC)
    # package_name (optional): return metrics for package_name
    start_date = request.args.get('start_date', None)
    end_date = request.args.get('end_date', None)
    package_name = request.args.get('package_name', None)

    if request.headers.get('Content-Type') == 'application/json':
        # This is the old, deprecated method of in-body parameters
        # We will keep it for backwards compatibility
        if start_date is None:
            start_date = request.json.get('start_date', None)
        if end_date is None:
            end_date = request.json.get('end_date', None)
        if package_name is None:
            package_name = request.json.get('package_name', None)

    if start_date is None or end_date is None:
        raise InvalidUsage('Missing parameters', status_code=400)

    # Convert dates to timestamp
    fmt = '%Y-%m-%d'
    try:
        start_timestamp = int(calendar.timegm(time.strptime(start_date, fmt)))
        end_timestamp = int(calendar.timegm(time.strptime(end_date, fmt)))
    except ValueError:
        raise InvalidUsage('Invalid date format, it must be YYYY-mm-dd',
                           status_code=400)

    # Find the commits count for each metric
    session = getSession(app.config['DB_PATH'])
    commits = session.query(Commit).filter(
        Commit.status == 'SUCCESS',
        Commit.dt_build >= start_timestamp,
        Commit.dt_build < end_timestamp)

    if package_name:
        commits = commits.filter(
            Commit.project_name == package_name)

    successful_commits = commits.count()

    commits = session.query(Commit).filter(
        Commit.status == 'FAILED',
        Commit.dt_build >= start_timestamp,
        Commit.dt_build <= end_timestamp)

    if package_name:
        commits = commits.filter(
            Commit.project_name == package_name)

    failed_commits = commits.count()
    total_commits = successful_commits + failed_commits

    result = {'succeeded': successful_commits,
              'failed': failed_commits,
              'total': total_commits}
    closeSession(session)
    return jsonify(result), 200


@app.route('/api/last_tested_repo', methods=['POST'])
@auth.login_required
@_json_media_type
def last_tested_repo_POST():
    # max_age: Maximum age in hours, used as base for the search
    # success(optional): find repos with a successful/unsuccessful vote
    # job_id(optional); name of the CI that sent the vote
    # reporting_job_id: name of the CI that will test this repo
    # sequential_mode(optional): if set to true, change the search algorithm
    #                            to only use previous_job_id as CI name to
    #                            search for. Defaults to false
    # previous_job_id(optional): CI name to search for, if sequential_mode is
    #                            True
    # component(optional): only get votes for this component
    max_age = request.json.get('max_age', None)
    my_job_id = request.json.get('reporting_job_id', None)
    job_id = request.json.get('job_id', None)
    success = request.json.get('success', None)
    sequential_mode = request.json.get('sequential_mode', None)
    previous_job_id = request.json.get('previous_job_id', None)
    component = request.json.get('component', None)

    if success is not None:
        success = bool(strtobool(success))

    if sequential_mode is not None:
        sequential_mode = bool(strtobool(sequential_mode))

    if sequential_mode and previous_job_id is None:
        raise InvalidUsage('Missing parameter previous_job_id',
                           status_code=400)

    if (max_age is None or my_job_id is None):
        raise InvalidUsage('Missing parameters', status_code=400)

    # Calculate timestamp as now - max_age
    if int(max_age) == 0:
        timestamp = 0
    else:
        oldest_time = datetime.now() - timedelta(hours=int(max_age))
        timestamp = time.mktime(oldest_time.timetuple())

    session = getSession(app.config['DB_PATH'])

    try:
        if sequential_mode:
            # CI pipeline case
            vote = getVote(session, timestamp, success, previous_job_id,
                           component=component, fallback=False)
        else:
            # Normal case
            vote = getVote(session, timestamp, success, job_id,
                           component=component)
    except Exception as e:
        raise e

    newvote = CIVote(commit_id=vote.commit_id, ci_name=my_job_id,
                     ci_url='', ci_vote=False, ci_in_progress=True,
                     timestamp=int(time.time()), notes='',
                     user=auth.username(), component=vote.component)
    session.add(newvote)
    session.commit()

    commit = session.query(Commit).filter(
        Commit.status == 'SUCCESS',
        Commit.id == vote.commit_id).first()

    result = {'commit_hash': commit.commit_hash,
              'distro_hash': commit.distro_hash,
              'timestamp': newvote.timestamp,
              'job_id': newvote.ci_name,
              'success': newvote.ci_vote,
              'in_progress': newvote.ci_in_progress,
              'user': newvote.user,
              'component': newvote.component}
    closeSession(session)
    return jsonify(result), 201


@app.route('/api/report_result', methods=['POST'])
@auth.login_required
@_json_media_type
def report_result():
    # job_id: name of CI
    # commit_hash: commit hash
    # distro_hash: distro hash
    # aggregate_hash: hash of aggregate.
    # url: URL where more information can be found
    # timestamp: CI execution timestamp
    # success: boolean
    # notes(optional): notes
    # Either commit_hash+distro_hash or aggregate_hash must be provided
    try:
        timestamp = request.json['timestamp']
        job_id = request.json['job_id']
        success = request.json['success']
        url = request.json['url']
    except KeyError:
        raise InvalidUsage('Missing parameters', status_code=400)

    commit_hash = request.json.get('commit_hash', None)
    distro_hash = request.json.get('distro_hash', None)
    aggregate_hash = request.json.get('aggregate_hash', None)

    if not commit_hash and not distro_hash and not aggregate_hash:
        raise InvalidUsage('Missing parameters', status_code=400)

    if commit_hash and not distro_hash:
        raise InvalidUsage('If commit_hash is provided, distro_hash '
                           'must be provided too', status_code=400)

    if distro_hash and not commit_hash:
        raise InvalidUsage('If distro_hash is provided, commit_hash '
                           'must be provided too', status_code=400)

    if (aggregate_hash and distro_hash) or (aggregate_hash and commit_hash):
        raise InvalidUsage('aggregate_hash and commit/distro_hash cannot be '
                           'combined', status_code=400)

    notes = request.json.get('notes', '')

    session = getSession(app.config['DB_PATH'])
    # We have two paths here: one for votes on commit/distro hash, another
    # for votes on aggregate_hash
    component = None
    if commit_hash:
        commit = _get_commit(session, commit_hash, distro_hash)
        if commit is None:
            raise InvalidUsage('commit_hash+distro_hash combination not found',
                               status_code=404)

        commit_id = commit.id
        component = commit.component
        vote = CIVote(commit_id=commit_id, ci_name=job_id, ci_url=url,
                      ci_vote=bool(strtobool(success)), ci_in_progress=False,
                      timestamp=int(timestamp), notes=notes,
                      user=auth.username(), component=component)
    else:
        prom = session.query(Promotion).filter(
            Promotion.aggregate_hash == aggregate_hash).first()
        if prom is None:
            raise InvalidUsage('aggregate_hash not found',
                               status_code=400)

        vote = CIVote_Aggregate(ref_hash=aggregate_hash, ci_name=job_id,
                                ci_url=url, ci_vote=bool(strtobool(success)),
                                ci_in_progress=False, timestamp=int(timestamp),
                                notes=notes, user=auth.username())

    session.add(vote)
    session.commit()

    result = {'commit_hash': commit_hash,
              'distro_hash': distro_hash,
              'aggregate_hash': aggregate_hash,
              'timestamp': timestamp,
              'job_id': job_id,
              'success': bool(strtobool(success)),
              'in_progress': False,
              'url': url,
              'notes': notes,
              'user': auth.username(),
              'component': component}
    closeSession(session)
    return jsonify(result), 201


@app.route('/api/promote', methods=['POST'])
@auth.login_required
@_json_media_type
def promote():
    # commit_hash: commit hash
    # distro_hash: distro hash
    # promote_name: symlink name
    try:
        commit_hash = request.json['commit_hash']
        distro_hash = request.json['distro_hash']
        promote_name = request.json['promote_name']
    except KeyError:
        raise InvalidUsage('Missing parameters', status_code=400)

    # Check for invalid promote names
    if (promote_name == 'consistent' or promote_name == 'current'):
        raise InvalidUsage('Invalid promote_name %s' % promote_name,
                           status_code=403)

    config_options = _get_config_options(app.config['CONFIG_FILE'])

    session = getSession(app.config['DB_PATH'])
    commit = _get_commit(session, commit_hash, distro_hash)
    if commit is None:
        raise InvalidUsage('commit_hash+distro_hash combination not found',
                           status_code=404)

    # If the commit has been purged, do not move on
    if commit.flags & FLAG_PURGED:
        raise InvalidUsage('commit_hash+distro_hash has been purged, cannot '
                           'promote it', status_code=410)

    if config_options.use_components:
        base_directory = os.path.join(app.config['REPO_PATH'], "component/%s" %
                                      commit.component)
    else:
        base_directory = app.config['REPO_PATH']

    target_link = os.path.join(base_directory, promote_name)
    # Check for invalid target links, like ../promotename
    target_dir = os.path.dirname(os.path.abspath(target_link))
    if not os.path.samefile(target_dir, base_directory):
        raise InvalidUsage('Invalid promote_name %s' % promote_name,
                           status_code=403)

    # We should create a relative symlink
    yumrepodir = commit.getshardedcommitdir()
    if config_options.use_components:
        # In this case, the relative path should not include
        # the component part
        yumrepodir = yumrepodir.replace("component/%s/" % commit.component, '')

    # Remove symlink if it exists, so we can create it again
    if os.path.lexists(os.path.abspath(target_link)):
        os.remove(target_link)
    try:
        os.symlink(yumrepodir, target_link)
    except Exception as e:
        raise InvalidUsage("Symlink creation failed with error: %s" %
                           e, status_code=500)

    # Once the updated symlink is created, if we are using components
    # we need to update the top-level repo file
    repo_checksum = None
    if config_options.use_components:
        datadir = os.path.realpath(config_options.datadir)
        repo_checksum = aggregate_repo_files(promote_name, datadir, session,
                                             config_options.reponame,
                                             hashed_dir=True)

    timestamp = time.mktime(datetime.now().timetuple())
    promotion = Promotion(commit_id=commit.id, promotion_name=promote_name,
                          timestamp=timestamp, user=auth.username(),
                          component=commit.component,
                          aggregate_hash=repo_checksum)

    session.add(promotion)
    session.commit()

    repo_hash = _repo_hash(commit)
    repo_url = "%s/%s" % (config_options.baseurl, commit.getshardedcommitdir())

    result = {'commit_hash': commit_hash,
              'distro_hash': distro_hash,
              'repo_hash': repo_hash,
              'repo_url': repo_url,
              'promote_name': promote_name,
              'component': commit.component,
              'timestamp': timestamp,
              'user': auth.username(),
              'aggregate_hash': repo_checksum}
    closeSession(session)
    return jsonify(result), 201


@app.route('/api/promote-batch', methods=['POST'])
@auth.login_required
@_json_media_type
def promote_batch():
    # hash_pairs: list of commit/distro hash pairs
    # promote_name: symlink name
    hash_list = []
    try:
        for pair in request.json:
            commit_hash = pair['commit_hash']
            distro_hash = pair['distro_hash']
            promote_name = pair['promote_name']
            hash_item = [commit_hash, distro_hash, promote_name]
            hash_list.append(hash_item)
    except KeyError:
        raise InvalidUsage('Missing parameters', status_code=400)

    config_options = _get_config_options(app.config['CONFIG_FILE'])
    session = getSession(app.config['DB_PATH'])
    # Now we will be running all checks for each combination
    # Check for invalid promote names
    for hash_item in hash_list:
        commit_hash = hash_item[0]
        distro_hash = hash_item[1]
        promote_name = hash_item[2]
        if (promote_name == 'consistent' or promote_name == 'current'):
            raise InvalidUsage('Invalid promote_name %s for hash %s_%s' % (
                               promote_name, commit_hash, distro_hash),
                               status_code=403)
        commit = _get_commit(session, commit_hash, distro_hash)
        if commit is None:
            raise InvalidUsage('commit_hash+distro_hash combination not found'
                               ' for %s_%s' % (commit_hash, distro_hash),
                               status_code=404)

        # If the commit has been purged, do not move on
        if commit.flags & FLAG_PURGED:
            raise InvalidUsage('commit_hash+distro_hash %s_%s has been purged,'
                               ' cannot promote it' % (commit_hash,
                                                       distro_hash),
                               status_code=410)

        if config_options.use_components:
            base_directory = os.path.join(app.config['REPO_PATH'],
                                          "component/%s" % commit.component)
        else:
            base_directory = app.config['REPO_PATH']

        target_link = os.path.join(base_directory, promote_name)
        # Check for invalid target links, like ../promotename
        target_dir = os.path.dirname(os.path.abspath(target_link))
        if not os.path.samefile(target_dir, base_directory):
            raise InvalidUsage('Invalid promote_name %s' % promote_name,
                               status_code=403)

    # After all checks have been performed, do all promotions
    rollback_list = []
    for hash_item in hash_list:
        rollback_item = {}
        commit_hash = hash_item[0]
        distro_hash = hash_item[1]
        promote_name = hash_item[2]
        commit = _get_commit(session, commit_hash, distro_hash)
        # We should create a relative symlink
        yumrepodir = commit.getshardedcommitdir()
        if config_options.use_components:
            base_directory = os.path.join(app.config['REPO_PATH'],
                                          "component/%s" %
                                          commit.component)
            # In this case, the relative path should not include
            # the component part
            yumrepodir = yumrepodir.replace("component/%s/" % commit.component,
                                            '')
        else:
            base_directory = app.config['REPO_PATH']

        target_link = os.path.join(base_directory, promote_name)
        rollback_item['target_link'] = target_link
        rollback_item['previous_link'] = None
        # Remove symlink if it exists, so we can create it again
        if os.path.lexists(os.path.abspath(target_link)):
            rollback_item['previous_link'] = os.readlink(
                os.path.abspath(target_link))
            os.remove(target_link)

        rollback_list.append(rollback_item)
        # This is the only destructive operation. If something fails here,
        # we will try to roll everything back
        try:
            os.symlink(yumrepodir, target_link)
        except Exception as e:
            _rollback_batch_promotion(rollback_list)
            raise InvalidUsage("Symlink creation failed with error: %s. "
                               "All previously created symlinks have been "
                               "rolled back." %
                               e, status_code=500)

        timestamp = time.mktime(datetime.now().timetuple())
        promotion = Promotion(commit_id=commit.id,
                              promotion_name=promote_name,
                              timestamp=timestamp, user=auth.username(),
                              component=commit.component,
                              aggregate_hash=None)
        session.add(promotion)

    # And finally, if we are using components, update the top-level
    # repo file
    repo_checksum = None
    if config_options.use_components:
        datadir = os.path.realpath(config_options.datadir)
        repo_checksum = aggregate_repo_files(promote_name, datadir, session,
                                             config_options.reponame,
                                             hashed_dir=True)
        promotion.aggregate_hash = repo_checksum
        session.add(promotion)

    # Close session and return the last promotion we did (which includes the
    # repo checksum)
    session.commit()
    repo_hash = _repo_hash(commit)
    repo_url = "%s/%s" % (config_options.baseurl, commit.getshardedcommitdir())
    result = {'commit_hash': commit_hash,
              'distro_hash': distro_hash,
              'repo_hash': repo_hash,
              'repo_url': repo_url,
              'promote_name': promote_name,
              'component': commit.component,
              'timestamp': timestamp,
              'user': auth.username(),
              'aggregate_hash': repo_checksum}
    closeSession(session)
    return jsonify(result), 201


@app.route('/api/remote/import', methods=['POST'])
@auth.login_required
@_json_media_type
def remote_import():
    # repo_url: repository URL to import from
    try:
        repo_url = request.json['repo_url']
    except KeyError:
        raise InvalidUsage('Missing parameters', status_code=400)

    try:
        import_commit(repo_url, app.config['CONFIG_FILE'],
                      db_connection=app.config['DB_PATH'])
    except Exception as e:
        raise InvalidUsage("Remote import failed with error: %s" %
                           e, status_code=500)

    result = {'repo_url': repo_url}
    return jsonify(result), 201


@app.template_filter()
def strftime(date, fmt="%Y-%m-%d %H:%M:%S"):
    gmdate = time.gmtime(date)
    return "%s" % time.strftime(fmt, gmdate)


@app.route('/api/civotes.html', methods=['GET'])
def get_civotes():
    session = getSession(app.config['DB_PATH'])
    offset = request.args.get('offset', 0)

    votes = session.query(CIVote)
    votes = votes.filter(CIVote.ci_name != 'consistent')
    votes = votes.order_by(desc(CIVote.timestamp))
    votes = votes.offset(offset).limit(pagination_limit)
    count = votes.count()
    # Let's find all individual commit_hash + distro_hash combinations
    commit_id_list = []
    for vote in votes:
        if vote.commit_id not in commit_id_list:
            commit_id_list.append(vote.commit_id)

    # Populate list for commits
    repolist = []
    for commit_id in commit_id_list:
        commit = getCommits(session, limit=0).filter(
            Commit.id == commit_id).first()

        repodetail = RepoDetail()
        repodetail.commit_hash = commit.commit_hash
        repodetail.distro_hash = commit.distro_hash
        repodetail.distro_hash_short = commit.distro_hash[:8]
        repodetail.success = votes.from_self().filter(
            CIVote.commit_id == commit_id, CIVote.ci_vote == 1).count()
        repodetail.failure = votes.from_self().filter(
            CIVote.commit_id == commit_id, CIVote.ci_vote == 0).count()
        repodetail.timestamp = votes.from_self().filter(
            CIVote.commit_id == commit_id).order_by(desc(CIVote.timestamp)).\
            first().timestamp
        repodetail.component = commit.component
        repolist.append(repodetail)

    repolist = sorted(repolist, key=lambda repo: repo.timestamp, reverse=True)

    closeSession(session)

    config_options = _get_config_options(app.config['CONFIG_FILE'])

    return render_template('votes_general.j2',
                           target=config_options.target,
                           repodetail=repolist,
                           count=count,
                           limit=pagination_limit)


@app.route('/api/civotes_detail.html', methods=['GET'])
def get_civotes_detail():
    commit_hash = request.args.get('commit_hash', None)
    distro_hash = request.args.get('distro_hash', None)
    component = request.args.get('component', None)
    ci_name = request.args.get('ci_name', None)
    success = request.args.get('success', None)
    offset = request.args.get('offset', 0)

    session = getSession(app.config['DB_PATH'])
    votes = session.query(CIVote)
    votes = votes.filter(CIVote.ci_name != 'consistent')

    if commit_hash and distro_hash:
        commit = _get_commit(session, commit_hash, distro_hash)
        votes = votes.from_self().filter(CIVote.commit_id == commit.id)
    elif ci_name:
        votes = votes.filter(CIVote.ci_name == ci_name)
    elif component:
        votes = votes.filter(CIVote.component == component)
    else:
        raise InvalidUsage("Please specify either commit_hash+distro_hash or "
                           "ci_name as parameters.", status_code=400)

    votes = votes.offset(offset).limit(pagination_limit)

    if success is not None:
        votes = votes.from_self().filter(
            CIVote.ci_vote == bool(strtobool(success)))

    votelist = votes.all()
    count = votes.count()

    for i in range(len(votelist)):
        commit = getCommits(session, limit=0).filter(
            Commit.id == votelist[i].commit_id).first()
        votelist[i].commit_hash = commit.commit_hash
        votelist[i].distro_hash = commit.distro_hash
        votelist[i].distro_hash_short = commit.distro_hash[:8]

    closeSession(session)
    config_options = _get_config_options(app.config['CONFIG_FILE'])

    return render_template('votes.j2',
                           target=config_options.target,
                           votes=votelist,
                           count=count,
                           limit=pagination_limit)


@app.route('/api/civotes_agg.html', methods=['GET'])
def get_civotes_agg():
    session = getSession(app.config['DB_PATH'])
    offset = request.args.get('offset', 0)

    votes = session.query(CIVote_Aggregate)
    votes = votes.order_by(desc(CIVote_Aggregate.timestamp))
    votes = votes.offset(offset).limit(pagination_limit)
    count = votes.count()
    # Let's find all individual aggregate_hashes
    agg_id_list = []
    for vote in votes:
        if vote.ref_hash not in agg_id_list:
            agg_id_list.append(vote.ref_hash)

    # Populate list for aggregates
    agglist = []
    for ref_hash in agg_id_list:
        aggdetail = AggDetail()
        aggdetail.ref_hash = ref_hash
        aggdetail.success = votes.from_self().filter(
            CIVote_Aggregate.ref_hash == ref_hash,
            CIVote_Aggregate.ci_vote == 1).count()
        aggdetail.failure = votes.from_self().filter(
            CIVote_Aggregate.ref_hash == ref_hash,
            CIVote_Aggregate.ci_vote == 0).count()
        aggdetail.timestamp = votes.from_self().filter(
            CIVote_Aggregate.ref_hash == ref_hash).\
            order_by(desc(CIVote_Aggregate.timestamp)).\
            first().timestamp
        agglist.append(aggdetail)

    agglist = sorted(agglist, key=lambda repo: repo.timestamp, reverse=True)

    closeSession(session)

    config_options = _get_config_options(app.config['CONFIG_FILE'])

    return render_template('votes_general_agg.j2',
                           target=config_options.target,
                           aggdetail=agglist,
                           count=count,
                           limit=pagination_limit)


@app.route('/api/civotes_agg_detail.html', methods=['GET'])
def get_civotes_agg_detail():
    ref_hash = request.args.get('ref_hash', None)
    ci_name = request.args.get('ci_name', None)
    success = request.args.get('success', None)
    offset = request.args.get('offset', 0)

    session = getSession(app.config['DB_PATH'])
    votes = session.query(CIVote_Aggregate)

    if ref_hash:
        votes = votes.from_self().filter(CIVote_Aggregate.ref_hash == ref_hash)
    elif ci_name:
        votes = votes.filter(CIVote_Aggregate.ci_name == ci_name)
    else:
        raise InvalidUsage("Please specify either ref_hash or "
                           "ci_name as parameters.", status_code=400)

    votes = votes.offset(offset).limit(pagination_limit)

    if success is not None:
        votes = votes.from_self().filter(
            CIVote_Aggregate.ci_vote == bool(strtobool(success)))

    votelist = votes.all()
    count = votes.count()

    closeSession(session)
    config_options = _get_config_options(app.config['CONFIG_FILE'])

    return render_template('votes_agg.j2',
                           target=config_options.target,
                           votes=votelist,
                           count=count,
                           limit=pagination_limit)


@app.route('/api/report.html', methods=['GET'])
def get_report():
    package_name = request.args.get('package', None)
    success = request.args.get('success', None)
    component = request.args.get('component', None)
    offset = request.args.get('offset', 0)

    if success is not None:
        if bool(strtobool(success)):
            with_status = "SUCCESS"
        else:
            with_status = "FAILED"
    else:
        with_status = None

    session = getSession(app.config['DB_PATH'])
    commits = getCommits(session, without_status="RETRY",
                         project=package_name, with_status=with_status,
                         limit=pagination_limit, offset=offset,
                         component=component)

    count = commits.count()

    config_options = _get_config_options(app.config['CONFIG_FILE'])
    closeSession(session)

    commits_build_dir = {}
    for commit in commits:
        version = ''
        release = ''
        commit_dir = commit.getshardedcommitdir()
        src_package = find_in_artifacts(commit.artifacts, r'\w+.src.rpm')
        if src_package:
            splitted_name = re.split(r'-0\.[0-9]{14}.[0-9a-f]{7}\.',
                                     src_package.split('/')[-1])
            # NOTE(dpawlik): Release is predictable, but versioning not.
            # So better is to take value after the split.
            version = splitted_name[0].replace(commit.project_name + '-', '')
            release = re.findall(r'0\.[0-9]{14}.[0-9a-f]{7}\.\w+',
                                 src_package.split('/')[-1])

        key = "%s_%s_%s" % (commit.commit_hash, commit.distro_hash,
                            commit.extended_hash)
        commits_build_dir[key] = {
            'build_dir': "%s/%s" % (config_options.baseurl, commit_dir),
            'version': version,
            'release': ''.join(release)
        }

    return render_template('report.j2',
                           reponame='Detailed build report',
                           target=config_options.target,
                           src=config_options.source,
                           project_name=config_options.project_name,
                           commits=commits,
                           count=count,
                           limit=pagination_limit,
                           commits_build_dir=commits_build_dir)
